from __future__ import annotations

from sqlalchemy.engine import Engine, create_engine
from sqlalchemy.orm import DeclarativeMeta, Session
from sqlalchemy_utils import database_exists

from apb.types import Pathish

# For now, this is a strictly SQLite3 library
DB_SCHEME = "sqlite:///"


def path_to_uri(path: Pathish) -> str:
    return f"{DB_SCHEME}{path}"


def engine_from_path(path: Pathish) -> Engine:
    uri = str(path)
    if not uri.startswith(DB_SCHEME):
        uri = path_to_uri(uri)

    return create_engine(uri)


def engine_from_session(session: Session) -> Engine:
    connection = session.get_bind()

    if not isinstance(connection, Engine):
        raise ValueError(f"Connection to {session} is not an engine")

    return connection


def create_db(engine: Engine, model: DeclarativeMeta, exist_ok: bool = False) -> None:
    """Create the database.

    Args:
        exist_ok:
            If False, raises an error if the database already exists. If True and the DB
            exists, no operation takes place.

    Raises:
        ValueError: If the database exists and `exist_ok` is False.
    """
    if database_exists(engine.url) and not exist_ok:
        raise ValueError("Database already exists.")

    model.metadata.create_all(engine)



import pandas as pd
import pytest

from apb.sequence.utils import aligned_seqs_to_dataframe


@pytest.fixture
def alignments() -> list[str]:
    return [
        "ATGC---ATGC",
        "ATGC---ATGC",
        "ATGCAAAATGC",
        "ATGCGGGATGC",
    ]


def test_aligned_seqs_to_frame(alignments):
    expected = pd.DataFrame(
        {
            0: ["A", "A", "A", "A"],
            1: ["T", "T", "T", "T"],
            2: ["G", "G", "G", "G"],
            3: ["C", "C", "C", "C"],
            4: ["-", "-", "A", "G"],
            5: ["-", "-", "A", "G"],
            6: ["-", "-", "A", "G"],
            7: ["A", "A", "A", "A"],
            8: ["T", "T", "T", "T"],
            9: ["G", "G", "G", "G"],
            10: ["C", "C", "C", "C"],
        }
    )

    # aligned_seqs_to_frame should produce expected
    assert aligned_seqs_to_dataframe(alignments).equals(expected)

    # Raises error if additional sequence has mismatched length
    with pytest.raises(ValueError, match="Not all sequences are of the same length"):
        # Add sequence that is too long
        alignments.append("ATGCGGGATGCATGCGGGATGCATGCGGGATGC")
        aligned_seqs_to_dataframe(alignments)



import json
import logging
from collections.abc import Mapping
from pathlib import Path
from typing import cast

import numpy as np
import pandas as pd
from dask.distributed import Client, Future
from numpy.typing import NDArray
from rich.progress import Progress

from apb.application.tmalign.core import ALLOWED_TMALIGN_EXTENSIONS, TMAlignOutput, run_tmalign
from apb.structure.utils import extension_matches
from apb.types import Pathish
from apb.utils import Timer, converter

# Dask & dask.distributed outputs more logging than we're interested in
logging.getLogger("dask").setLevel(logging.ERROR)
logging.getLogger("distributed").setLevel(logging.ERROR)

ManyToManyOutput = dict[str, dict[str, TMAlignOutput]]


def _build_from_inverse(inverse: TMAlignOutput) -> TMAlignOutput:
    return TMAlignOutput(
        qseq_aligned=inverse.tseq_aligned,
        tseq_aligned=inverse.qseq_aligned,
        qtmscore=inverse.ttmscore,
        ttmscore=inverse.qtmscore,
        rmsd=inverse.rmsd,
    )


def _validate_paths(paths: list[Pathish]) -> None:
    if not all(extension_matches(path, ALLOWED_TMALIGN_EXTENSIONS) for path in paths):
        raise ValueError(f"TMalign only accepts the extensions: {ALLOWED_TMALIGN_EXTENSIONS}")


def _submit_batch(
    batch: list[tuple[str, str]],
    client: Client,
    results: ManyToManyOutput,
    query_structures: Mapping[str, Pathish],
    target_structures: Mapping[str, Pathish],
) -> dict[str, dict[str, Future]]:
    """Submit a batch of TMalign jobs.

    The TMalign jobs are submitted in batches to avoid overloading the scheduler.
    Anecdotally, dask slows down substantially when the scheduler holds >500k jobs
    concurrently, even if the jobs are largely independent of one another. Understanding
    the circumstances of the slow-down would be useful, but batching is good practice in
    general and alleviates the problem.

    Args:
        batch:
            A list of two-ples, where the first and second elements are the
            structure identifiers of the query and target, respectively.
    """
    futures: dict[str, dict[str, Future]] = {}
    for query, target in batch:
        if query not in futures:
            futures[query] = {}

        # In what follows, we check if the inverse alignment has been calculated or
        # submitted. Since the alignment of query to target can be derived from the
        # alignment of target to query (see _build_from_inverse to see how), we can
        # build the alignment from its inverse, assuming the inverse has been or
        # will be calculated.

        if target in results and query in results[target]:
            # The inverse is in `results`, meaning it's already been calculated.
            future = client.submit(_build_from_inverse, results[target][query])
        elif target in futures and query in futures[target]:
            # The inverse is in `futures`, meaning it's been submitted. The job may
            # or may not be complete, but either way, we pass the `Future` object to
            # `client.submit` and the alignment will be built from its inverse when
            # ready.
            future = client.submit(_build_from_inverse, futures[target][query])
        else:
            future = client.submit(run_tmalign, query_structures[query], target_structures[target])

        futures[query][target] = future

    return futures


def align_many_to_many(
    query_structures: Mapping[str, Pathish],
    target_structures: Mapping[str, Pathish],
    client: Client,
    batch_size: int = 2000,
) -> ManyToManyOutput:
    """Perform many-to-many structure alignments using TMalign.

    This function computes all pairwise alignments between structures in
    `query_structures` and `target_structures` using TMalign.

    The computation is parallelized with Dask. If a Dask client is not provided, a new
    client will be created for the duration of the computation.

    The computation is also sped up by leveraging that TMalign has an inverse
    relationship with its inputs, whereby the alignment of query to target can be
    derived from the inverse alignment (target to query). By building alignments
    from their inverses, where possible, the amount of explicit calculation required is
    reduced by as much as half depending on the amount of query/target overlap.

    Args:
        query_structures:
            A mapping from structure identifiers to file paths for the query structures.
        target_structures:
            A mapping from structure identifiers to file paths for the target
            structures.
        client:
            A Dask client to use for computation. If None, a new client will be created.
            Defaults to None.
        batch_size:
            The number of alignments to submit per batch.

    Returns:
        dict[str, dict[str, TMAlignOutput]]:
            A nested dictionary containing the alignment results. The outer dictionary
            is keyed by query structure identifiers, and the inner dictionary is keyed
            by target structure identifiers. Each value is a `TMAlignOutput` object
            containing the alignment results.

    Raises:
        ValueError:
            If any of the file paths do not have an extension expected by TMalign.
    """

    # Vet all the paths prior to starting (many-to-many could be an expensive op).
    _validate_paths(list(query_structures.values()) + list(target_structures.values()))

    logging.info(f"View status: {client.dashboard_link}")
    logging.info(f"Progress updates every {batch_size} comparisons")

    results: ManyToManyOutput = {}
    num_comparisons = len(query_structures) * len(target_structures)
    total_batches = (num_comparisons + batch_size - 1) // batch_size
    query_target_pairs = [
        (query, target) for query in query_structures for target in target_structures
    ]

    with Progress(transient=True) as progress, Timer(quiet=True) as timer:
        task = progress.add_task(description="", total=num_comparisons)

        for batch_num in range(total_batches):
            count = batch_num * batch_size

            progress.update(task, description=f"Performing alignments [{count}/{num_comparisons}]")

            batch = query_target_pairs[count : count + batch_size]
            futures = _submit_batch(batch, client, results, query_structures, target_structures)

            # `client.gather` returns the same type passed to it, except with all `Future`
            # objects replaced with their results. We inform our type checker of this by
            # casting the type explictly.
            batch_results = cast(ManyToManyOutput, client.gather(futures))

            # Populate `results` with the results of the batch.
            for query, target_results in batch_results.items():
                results.setdefault(query, {}).update(target_results)

            progress.advance(task, len(batch))

    logging.info(f"{num_comparisons} alignments calculated in {timer.time}")

    return results


def load_alignments_json(path: Pathish) -> ManyToManyOutput:
    with open(path) as fp:
        return converter.structure(json.load(fp), ManyToManyOutput)


def save_alignments_json(path: Pathish, alignments: ManyToManyOutput) -> Path:
    with open(path, "w") as fp:
        json.dump(converter.unstructure(alignments), fp, indent=2)

    return Path(path)


def align_to_reference(
    reference_structure: Pathish,
    target_structures: Mapping[str, Pathish],
    client: Client,
) -> pd.DataFrame:
    """Performs alignment of a collection of structures to a reference structure using TMalign.

    Executes a pairwise alignment to the reference structure for each target structure
    and returns the alignment trace.

    Args:
        reference_structure: Path to the reference structure file.
        target_structures:
            A mapping from structure identifiers to file paths for the target structures.

    Returns:
        pd.DataFrame:
            A dataframe of the alignment trace. Each column is a target hit to the
            reference. The indices are the reference residue indices. For further
            details, see `apb.utils.StructuralReferenceAligner`.
    """
    alignments = align_many_to_many({"ref": reference_structure}, target_structures, client)["ref"]

    traces: dict[str, NDArray[np.int64]] = {}
    for target, alignment in alignments.items():
        biotite_alignment = alignment.to_biotite()
        target_trace = biotite_alignment.trace[:, 1]
        reference_trace = biotite_alignment.trace[:, 0]

        # Remove all align indices in the target where the reference is gapped.
        target_trace = target_trace[reference_trace != -1]

        traces[target] = target_trace

    return pd.DataFrame(traces)



from apb.database.collection.interface import (
    add_preprocessed_profiles,
    add_profiles,
    export_structures,
    get_engine,
    get_profile_from_index,
    get_profile_from_name,
    iter_all_profiles,
    load_all_profiles,
    preprocess_profile,
)
from apb.database.collection.schema import Protein, Residue, ResiduePair, residue_dataframe_schema

__all__ = [
    "add_profiles",
    "add_preprocessed_profiles",
    "get_profile_from_index",
    "get_profile_from_name",
    "preprocess_profile",
    "get_engine",
    "Protein",
    "Residue",
    "ResiduePair",
    "residue_dataframe_schema",
    "export_structures",
    "iter_all_profiles",
    "load_all_profiles",
]



"""A minimalistic Foldseek wrapper

This module provides a simplified interface to some of the basic functionalities of
Foldseek. It facilitates the use of Foldseek by abstracting the command-line
interactions.

**Important Note** This is a bare-bones implementation intended for basic use cases. It
does not cover all the flags, options, or outputs available in Foldseek. The goal is to
provide a starting point that can be extended and refined as use cases develop. This is
not and will never be a feature-complete wrapper of Foldseek. Nevertheless, users
requiring advanced features or customizations can consider extending the code to meet
their use case, or make a feature request.
"""

from pathlib import Path

import pandas as pd

from apb.application.foldseek.datatypes import (
    FoldSeekConvertalisConfig,
    FoldseekOutputFileNotFoundError,
    FoldSeekSearchConfig,
)
from apb.application.foldseek.utils import format_alignment_tsv
from apb.types import Pathish
from apb.utils import maybe_temp_directory, require_dependency, run_command

FOLDSEEK_EXECUTABLE: str = "foldseek"
LOG_FOLDSEEK_OUTPUT: bool = False
DEFAULT_SEARCH_CONFIG = FoldSeekSearchConfig()
DEFAULT_CONVERTALIS_CONFIG = FoldSeekConvertalisConfig()


@require_dependency(FOLDSEEK_EXECUTABLE)
def createdb(input_path: Pathish, output_path_prefix: Pathish) -> Path:
    """Runs foldseek's createdb subcommand.

    This runs the command `foldseek createdb <input_path> <output_path_prefix>` and
    checks that the expected output path exists. No other logic is layered in. Consult
    `foldseek createdb -h` for documentation. If any flags are needed, please request
    them.

    Returns:
        Path:
            The path to the main DB file. `foldseek createdb` generates many files, all
            prefixed with this path.
    """
    input_path = Path(input_path)
    output_path_prefix = Path(output_path_prefix)

    command = [
        FOLDSEEK_EXECUTABLE,
        "createdb",
        str(input_path),
        str(output_path_prefix),
    ]
    run_command(command, log_stdout=LOG_FOLDSEEK_OUTPUT)

    if not output_path_prefix.exists():
        raise FoldseekOutputFileNotFoundError(output_path_prefix, command)

    return output_path_prefix


@require_dependency(FOLDSEEK_EXECUTABLE)
def search(
    query_path: Pathish,
    target_path: Pathish,
    alignment_path_prefix: Pathish,
    config: FoldSeekSearchConfig = DEFAULT_SEARCH_CONFIG,
) -> Path:
    """Runs foldseek's search subcommand.

    This runs the base command:

    `foldseek search <query_path> <target_path> <alignment_path_prefix> tmp`

    Additionally, some parameters can be appended to this command by specifying them
    with a `FoldSeekSearchConfig` object.

    Args:
        query_path: The input query file path.
        target_path: The input target file path.
        alignment_path_prefix: The output prefix for the alignment files.
        config:
            Configuration for the search command. If not supplied, foldseek's default
            options are used.

    Returns:
        Path: The path prefix to the alignment files.
    """
    query_path = Path(query_path)
    target_path = Path(target_path)
    alignment_path_prefix = Path(alignment_path_prefix)

    command = [
        FOLDSEEK_EXECUTABLE,
        "search",
        str(query_path),
        str(target_path),
        str(alignment_path_prefix),
        "tmp",
    ]
    command.extend(config.parameters())
    run_command(command, log_stdout=LOG_FOLDSEEK_OUTPUT)

    # The prefix doesn't itself exist, so we check for the existence of the index.
    if not alignment_path_prefix.with_suffix(".index").exists():
        raise FoldseekOutputFileNotFoundError(alignment_path_prefix, command)

    return alignment_path_prefix


@require_dependency(FOLDSEEK_EXECUTABLE)
def convertalis(
    query_path: Pathish,
    target_path: Pathish,
    alignment_path_prefix: Pathish,
    tsv_path: Pathish | None = None,
    config: FoldSeekConvertalisConfig = DEFAULT_CONVERTALIS_CONFIG,
    working_dir: Pathish | None = None,
) -> pd.DataFrame:
    """Runs foldseek's convertalis subcommand and formats the alignment TSV.

    This runs the command:

    foldseek convertalis <query_path> <target_path> <alignment_path_prefix> <tsv_path> \
        --format-mode 4 --format-output <format_output>

    `--format-mode 4` (BLAST-TAB + column headers) is hardcoded as the only possible
    output mode.

    Args:
        query_path: The input query file path.
        target_path: The input target file path.
        alignment_path_prefix: The alignment path prefix.
        tsv_path: The output TSV file path. If not provided, a temporary file will be used.
        format_output:
            An optional format string to specify which columns should be present in the
            output. For available options see `foldseek convertalis -h`.

    Returns:
        pd.DataFrame: The formatted dataframe containing alignment results.
    """
    query_path = Path(query_path)
    target_path = Path(target_path)
    alignment_path_prefix = Path(alignment_path_prefix)

    with maybe_temp_directory(working_dir) as working_dir:
        tsv_path = Path(tsv_path) if tsv_path else working_dir / "alignment.tsv"

        to_tsv_command = [
            FOLDSEEK_EXECUTABLE,
            "convertalis",
            str(query_path),
            str(target_path),
            str(alignment_path_prefix),
            str(tsv_path),
            "--format-mode",
            "4",
        ]
        to_tsv_command.extend(config.parameters())

        run_command(to_tsv_command, log_stdout=LOG_FOLDSEEK_OUTPUT)

        if not tsv_path.exists():
            raise FoldseekOutputFileNotFoundError(alignment_path_prefix, to_tsv_command)

        return format_alignment_tsv(pd.read_csv(tsv_path, sep="\t"))



"""Utilities for alignment"""

import pandas as pd


def aligned_seqs_to_dataframe(seqs: list[str]) -> pd.DataFrame:
    """Convert a list of aligned (gapped) sequences to a dataframe

    Args:
        seqs: A list of strings, where each string is a gapped sequence of characters.

    Returns:
        frame:
            A DataFrame where each row corresponds to a sequence from the input list and
            each column represents a position in the sequence alignment.

    Raises:
        ValueError:
            If not all sequences in the input list are of the same length.
    """
    if not all(len(seq) == len(seqs[0]) for seq in seqs):
        raise ValueError("Not all sequences are of the same length")

    return pd.DataFrame([list(seq) for seq in seqs])



"""Vendored StrEnum for compatibility with Python <3.11.

Notes:
    - Source: https://github.com/python/cpython/blob/63b5aed06ed6fcd7956c9b99eb0a3651116cf2e4/Lib/enum.py#L1351
    - Types have been added.
    - Concise resource on the case for StrEnum: https://tsak.dev/posts/python-enum/.
"""

from enum import Enum, auto
from typing import Any, TypeVar

__all__ = ["auto"]

_S = TypeVar("_S", bound="StrEnum")


class StrEnum(str, Enum):
    """
    Enum where members are also (and must be) strings
    """

    def __new__(cls: type[_S], *values: str) -> _S:
        if len(values) > 3:
            raise TypeError("too many arguments for str(): %r" % (values,))
        if len(values) == 1:
            # it must be a string
            if not isinstance(values[0], str):
                raise TypeError("%r is not a string" % (values[0],))
        if len(values) >= 2:
            # check that encoding argument is a string
            value = values[1]  # type: ignore
            if not isinstance(value, str):
                raise TypeError("encoding must be a string, not %r" % (value,))
        if len(values) == 3:
            # check that errors argument is a string
            if not isinstance(values[2], str):
                raise TypeError("errors must be a string, not %r" % (values[2]))
        value = str(*values)
        member = str.__new__(cls, value)
        member._value_ = value
        return member

    __str__ = str.__str__  # type: ignore

    @staticmethod
    def _generate_next_value_(name: str, start: int, count: int, last_values: list[Any]) -> str:
        """
        Return the lower-cased version of the member name.
        """
        return name.lower()



import gzip
from io import BytesIO
from pathlib import Path

import biotite.structure.io as structio
import biotite.structure.io.pdb as pdb
from biotite.structure.io.pdbx import BinaryCIFFile, get_structure, set_structure

from apb.structure.utils import RECOGNIZED_STRUCTURE_FILETYPES, structure_name_and_extension
from apb.types import Pathish, ProteinStructure


def save_structure(path: Pathish, structure: ProteinStructure) -> Path:
    path = Path(path)

    _, ext = structure_name_and_extension(path)
    if ext is None:
        raise ValueError(
            f"Unsupported file format {path}. Supported extensions are: "
            f"{RECOGNIZED_STRUCTURE_FILETYPES}"
        )

    if ext == ".pdb.gz":
        # Biotite doesn't manage gzip ops, so we do it ourselves
        with gzip.open(path, "wt") as gz_file:
            pdbfile = pdb.PDBFile()
            pdb.set_structure(pdbfile, structure)
            pdbfile.write(gz_file)
    else:
        # Otherwise delegate to biotite's high-level IO API
        structio.save_structure(path, structure)

    return path


def load_structure(path: Pathish) -> ProteinStructure:
    path = Path(path)

    _, ext = structure_name_and_extension(path)
    if ext is None:
        raise ValueError(
            f"Unsupported file format {path}. Supported extensions are: "
            f"{RECOGNIZED_STRUCTURE_FILETYPES}"
        )

    if ext == ".pdb.gz":
        # Biotite doesn't manage gzip ops, so we do it ourselves
        with gzip.open(path, "rt") as gz_file:
            pdbfile = pdb.PDBFile.read(gz_file)
            structure = pdb.get_structure(pdbfile, model=1)
    else:
        # Otherwise delegate to biotite's high-level IO API
        structure = structio.load_structure(path)

    assert isinstance(structure, ProteinStructure)
    return structure


def structure_to_bytes(structure: ProteinStructure) -> bytes:
    with BytesIO() as buffer:
        bcif = BinaryCIFFile()
        set_structure(bcif, structure)
        bcif.write(buffer)
        return buffer.getvalue()


def structure_from_bytes(byte_representation: bytes) -> ProteinStructure:
    with BytesIO(byte_representation) as buffer:
        file = BinaryCIFFile.read(buffer)
        structure = get_structure(file, model=1)
        assert isinstance(structure, ProteinStructure)
        return structure



import pandera as pa
from sqlalchemy import (
    Column,
    Float,
    ForeignKey,
    Integer,
    LargeBinary,
    String,
)
from sqlalchemy.orm import declarative_base, relationship

from apb.types import aa_dtype, ss3_dtype

Base = declarative_base()


class Protein(Base):
    __tablename__ = "proteins"
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, index=True, unique=True)
    sequence = Column(String)
    structure = Column(LargeBinary)
    source = Column(String)

    residues = relationship("Residue", back_populates="protein")
    residue_pairs = relationship("ResiduePair", back_populates="protein")


class Residue(Base):
    __tablename__ = "residues"
    id = Column(Integer, primary_key=True, autoincrement=True)
    protein_id = Column(Integer, ForeignKey("proteins.id"), index=True)
    residue_index = Column(Integer)
    amino_acid = Column(String)
    secondary_structure = Column(String)
    solvent_accessibility = Column(Float)
    relative_solvent_accessibility = Column(Float)

    protein = relationship("Protein", back_populates="residues")
    residue_pair1 = relationship(
        "ResiduePair",
        foreign_keys="[ResiduePair.residue1_id]",
        back_populates="residue1",
        overlaps="residue_pair2,residue1",
    )
    residue_pair2 = relationship(
        "ResiduePair",
        foreign_keys="[ResiduePair.residue2_id]",
        back_populates="residue2",
        overlaps="residue_pair1,residue2",
    )


class ResiduePair(Base):
    """Residue-to-residue measures

    This table holds information related to residue pairs. The most canonical example
    is the Euclidean distance between two residues.

    TODO currently unused
    """

    __tablename__ = "residue_pairs"
    id = Column(Integer, primary_key=True, autoincrement=True)
    protein_id = Column(Integer, ForeignKey("proteins.id"), index=True)
    residue1_id = Column(Integer, ForeignKey("residues.id"))
    residue2_id = Column(Integer, ForeignKey("residues.id"))
    distance = Column(Float)

    protein = relationship("Protein", back_populates="residue_pairs")
    residue1 = relationship("Residue", foreign_keys=[residue1_id], uselist=False)
    residue2 = relationship("Residue", foreign_keys=[residue2_id], uselist=False)


residue_dataframe_schema = pa.DataFrameSchema(
    {
        Residue.residue_index.key: pa.Column("uint32"),
        Residue.amino_acid.key: pa.Column(aa_dtype),
        Residue.secondary_structure.key: pa.Column(
            ss3_dtype,
            nullable=True,
        ),
        Residue.solvent_accessibility.key: pa.Column(
            "float64",
            checks=[pa.Check.ge(0.0)],
            nullable=True,
        ),
        Residue.relative_solvent_accessibility.key: pa.Column(
            "float64",
            checks=[pa.Check.le(1.0), pa.Check.ge(0.0)],
            nullable=True,
        ),
    },
    coerce=True,
)



from __future__ import annotations
import logging
from collections.abc import Mapping
from functools import partial

import click
import pandas as pd
import xarray as xr
from dask.distributed import Client, LocalCluster
from sqlalchemy.orm import Session

import apb.application.foldseek as foldseek
import apb.application.tmalign as tmalign
import apb.database.collection as collection
from apb.structure.convert import save_structure
from apb.structure.utils import structure_name_and_extension
from apb.types import Pathish, StructuralReferenceAligner, not_none
from apb.utils import maybe_temp_directory


def create_feature_alignment(
    aligner: StructuralReferenceAligner,
    collection_db: Pathish,
    reference_structure: Pathish,
    target_structures: Mapping[str, Pathish],
) -> xr.Dataset:
    """Creates an alignment of features from a protein collection to a reference protein.

    Implementation details: This function aligns each target protein to the reference
    protein using `aligner`. From the alignment traces, the residue features of each
    protein are mapped to the reference protein indices. This is done by creating a
    dataframe of the residue features, where the index corresponds to the indices of the
    reference protein instead of the indices of the original protein. All these pairwise
    feature alignment frames are then concatenated into a multi-index dataframe and
    finally converted into an xarray dataset.

    Args:
        aligner: The aligner object used for alignment.
        collection_db: Path to the collection database.
        reference_structure: Path to the reference structure file.
        target_structures: Dictionary with keys as names and values as structure paths.

    Returns:
        xr.Dataset: The resulting xarray dataset containing the aligned features.
    """
    traces = aligner(reference_structure, target_structures)

    aligned_frames: dict[str, pd.DataFrame] = {}

    with Session(collection.get_engine(collection_db)) as session:
        for profile in collection.iter_all_profiles(session):
            if profile.name not in traces.columns:
                # The protein didn't have a hit to the reference
                continue

            # Index is the protein trace, values are the reference trace
            ref_index = pd.Series(traces.index, index=traces[profile.name].values)

            # Drop any deletions relative to the reference
            ref_index = ref_index.loc[ref_index.index != -1]

            # Add ref_index to a copy of profile.residue_features. Now there is a
            # relationship between each residue feature and the reference index
            aligned_frame = profile.residue_features.copy()
            aligned_frame["ref_index"] = ref_index
            aligned_frame["ref_index"] = aligned_frame["ref_index"].astype("UInt32")

            # Removes insertions relative to the reference
            aligned_frame = aligned_frame.dropna(subset=["ref_index"])

            # Currently the index is the protein index. Swap the protein index with the
            # reference index
            aligned_frame = aligned_frame.reset_index().set_index("ref_index")

            aligned_frames[profile.name] = aligned_frame

    # Concatenate all the aligned frames into a multi-index, where each index specifies
    # the reference index (`ref_index`) for a given protein (`name`) residue. Note that
    # since not every alignment contains every reference index, these dataframes are all
    # of different shape, but share the same index superset: the reference indices.
    aligned_multiindex_frame = pd.concat(aligned_frames, names=["name", "ref_index"])

    # Convert to an xarray dataset. Each index level becomes a dimension, so the dataset
    # has dimensions `name` and `ref_index`. Each residue feature becomes an xarray
    # variable. The variables are defined for each reference index in every protein,
    # meaning any missing data is filled with null values. For the `amino_acid`
    # variable, we set these null values to the gap character, `-`.
    dataset = aligned_multiindex_frame.to_xarray()
    dataset["amino_acid"] = dataset["amino_acid"].fillna("-")

    return dataset


def get_preconfigured_aligners(client: Client) -> dict[str, StructuralReferenceAligner]:
    """Return the structural aligners used for aligning to reference."""
    return {
        "tmalign": partial(tmalign.align_to_reference, client=client),
        "foldseek_tmalign": foldseek.create_reference_aligner(
            search_config=foldseek.FoldSeekSearchConfig(
                backtrace=True,
                # Align the structures using Foldseek's version of the TMalign
                # algorithm (alignment_type=1). This is in contrast to doing a
                # sequence alignment of the structures' 3Di sequences. This is the
                # slowest and most accurate align option in Foldseek.
                alignment_type=1,
                # Do not perform pre-filtering (prefilter_mode=2). Pre-filtering
                # avoids comparisons of some proteins based on kmer frequencies of
                # the 3Di sequences.
                prefilter_mode=2,
                # Do not take shortcuts with TMalign-based alignment.
                tmalign_fast=0,
            ),
        ),
        "foldseek_tmalign_fast": foldseek.create_reference_aligner(
            search_config=foldseek.FoldSeekSearchConfig(
                backtrace=True,
                # Align the structures using Foldseek's version of the TMalign
                # algorithm (alignment_type=1). This is in contrast to doing a
                # sequence alignment of the structures' 3Di sequences. This is the
                # slowest and most accurate align option in Foldseek.
                alignment_type=1,
                # Do not perform pre-filtering (prefilter_mode=2). Pre-filtering
                # avoids comparisons of some proteins based on kmer frequencies of
                # the 3Di sequences.
                prefilter_mode=2,
                # Take shortcuts with the TMalign-based alignment.
                tmalign_fast=1,
            ),
        ),
        "foldseek_3di": foldseek.create_reference_aligner(
            search_config=foldseek.FoldSeekSearchConfig(
                backtrace=True,
                # Align the structures using Foldseek's 3Di+AA alignment option.
                # This is an alignment based on both the 3Di sequences and the AA
                # sequences. This is the default Foldseek option.
                alignment_type=2,
                # Do not perform pre-filtering (prefilter_mode=2). Pre-filtering
                # avoids comparisons of some proteins based on kmer frequencies of
                # the 3Di sequences.
                prefilter_mode=2,
            ),
        ),
    }


def run(
    collection_db: Pathish,
    reference_name: str,
    aligners: dict[str, StructuralReferenceAligner],
    client: Client,
    output: Pathish,
):
    with maybe_temp_directory() as tmpdir:
        structure_extension = ".pdb"
        reference_structure_path = tmpdir / f"{reference_name}{structure_extension}"
        target_structures_dir = tmpdir / "targets"

        logging.info("Saving structures to temporary directory")
        with Session(collection.get_engine(collection_db)) as session:
            reference_profile = collection.get_profile_from_name(session, reference_name)
            save_structure(reference_structure_path, reference_profile.structure)
            collection.export_structures(session, target_structures_dir, client)

        target_structures = {
            structure_name_and_extension(path)[0] | not_none: path
            for path in target_structures_dir.glob(f"*{structure_extension}")
        }

        # We are going to calculate a residue feature dataset for each method and
        # concatenate them along a dimension named "align_method". `datasets` will
        # hold the dataset for each method.
        datasets = []

        for method, aligner in aligners.items():
            logging.info(f"Running alignment with {method=}")

            datasets.append(
                create_feature_alignment(
                    aligner,
                    collection_db,
                    reference_structure_path,
                    target_structures,
                )
            )

            logging.info(f"Alignment dataset saved to {output}")

    combined_dataset = xr.concat(datasets, dim=pd.Index(aligners.keys(), name="align_method"))
    combined_dataset.to_netcdf(output)


@click.command()
@click.option(
    "-c",
    "--collection-db",
    required=True,
    type=click.Path(),
    help="Path to the collection database.",
)
@click.option(
    "-r",
    "--reference-name",
    required=True,
    type=str,
    help="Name of the reference structure.",
)
@click.option(
    "-o",
    "--output",
    required=True,
    type=click.Path(),
    help="Output NetCDF file.",
)
def main(
    collection_db: Pathish,
    reference_name: str,
    output: Pathish,
):
    with LocalCluster() as cluster, Client(cluster) as client:
        logging.info(f"Dask client: {client}")
        logging.info(f"View status at {client.dashboard_link}")

        preconfigured_aligners = get_preconfigured_aligners(client)
        run(collection_db, reference_name, preconfigured_aligners, client, output)


if __name__ == "__main__":
    main()



from collections.abc import Iterator
from pathlib import Path

import attrs
import numpy as np
import pandas as pd
import pytest
from biotite.sequence.seqtypes import ProteinSequence
from biotite.structure.residues import get_residues
from dask.distributed import Client, LocalCluster
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session
from sqlalchemy_utils import database_exists

import apb.database.collection as collection
from apb._test_pointer import structure_paths
from apb.database.collection.interface import (
    export_structures,
    iter_all_profiles,
    load_all_profiles,
)
from apb.database.collection.schema import Residue, residue_dataframe_schema
from apb.database.utils import DB_SCHEME, path_to_uri
from apb.profile import ProteinProfile
from apb.structure.convert import load_structure
from apb.structure.features.accessibility import calc_solvent_accessibility
from apb.structure.features.accessibility.datatypes import SolventAccessibilityConfig
from apb.structure.features.secondary_structure import calc_3_state_secondary_structure
from apb.structure.utils import (
    RECOGNIZED_STRUCTURE_FILETYPES,
    structure_name_and_extension,
    structures_almost_equal,
)
from apb.types import not_none
from apb.utils import maybe_temp_directory


@pytest.fixture
def db(tmp_path):
    return collection.get_engine(tmp_path / "test.db", create=True)


@pytest.fixture
def profiles() -> list[ProteinProfile]:
    profiles: list[ProteinProfile] = []

    for path in structure_paths():
        structure = load_structure(path)
        name = structure_name_and_extension(path)[0] | not_none
        sequence = ProteinSequence(get_residues(structure)[1])

        rsa_config = attrs.evolve(SolventAccessibilityConfig.default(), normalize=True)
        sasa_config = attrs.evolve(SolventAccessibilityConfig.default(), normalize=False)

        res_features = pd.DataFrame(
            {
                Residue.secondary_structure.key: calc_3_state_secondary_structure(structure),
                Residue.solvent_accessibility.key: calc_solvent_accessibility(
                    structure, sasa_config
                ),
                Residue.relative_solvent_accessibility.key: calc_solvent_accessibility(
                    structure, rsa_config
                ),
            }
        )

        # When features don't have a value for each residue, their missing values in the
        # `ProteinProfile.residue_features` are cast as null values. It's imperative
        # that missing data maintain datatype consistency as they are written to and
        # subsequently read from a collection DB. This is simulated by randomly setting
        # 10 of any nullable feature's values to null.
        for feature_name in res_features.columns:
            if residue_dataframe_schema.columns[feature_name].nullable:
                indices_to_null = np.random.choice(res_features.shape[0], 10, replace=False)
                res_features.loc[indices_to_null, feature_name] = None

        profiles.append(
            ProteinProfile(
                name=name,
                sequence=sequence,
                structure=structure,
                residue_features=res_features,
            )
        )

    return profiles


def test_load_db(tmp_path: Path):
    # This DB doesn't exist
    db_path = tmp_path / "test.db"
    assert not db_path.exists()
    assert not database_exists(path_to_uri(db_path))

    # Problem loading non-existant DB
    with pytest.raises(ValueError):
        return collection.get_engine(db_path, create=False)

    # It still doesn't exist (no side effects from the call)
    assert not db_path.exists()
    assert not database_exists(path_to_uri(db_path))

    # No problem if create is set to True
    collection.get_engine(db_path, create=True)

    # Now it exists
    assert db_path.exists()
    assert database_exists(path_to_uri(db_path))


def test_get_profile(db: Engine, profiles: list[ProteinProfile]):
    profile = profiles[0]

    # Fetching profile from empty DB doesn't work
    with Session(db) as session:
        with pytest.raises(KeyError, match="is not present"):
            collection.get_profile_from_index(session, 1)

    with Session(db) as session:
        with pytest.raises(KeyError, match="is not present"):
            collection.get_profile_from_name(session, profile.name)

    # Write the profile to the DB
    with Session(db) as session:
        collection.add_profiles(session, (profile,))

    # Fetching profile using non-existant keys doesn't work.
    with Session(db) as session:
        with pytest.raises(KeyError, match="is not present"):
            collection.get_profile_from_index(session, 2)

    with Session(db) as session:
        with pytest.raises(KeyError, match="is not present"):
            collection.get_profile_from_name(session, "foo")

    # Fetching the profile with the correct keys works
    with Session(db) as session:
        collection.get_profile_from_index(session, 1)

    with Session(db) as session:
        collection.get_profile_from_name(session, profile.name)


def _assert_profiles_equal(profile: ProteinProfile, other: ProteinProfile) -> None:
    assert isinstance(profile, ProteinProfile)
    assert isinstance(other, ProteinProfile)
    assert profile.name == other.name
    assert structures_almost_equal(profile.structure, other.structure)
    assert profile.sequence == other.sequence
    assert other.residue_features.equals(profile.residue_features)
    assert (other.residue_features.dtypes == profile.residue_features.dtypes).all()


def test_get_profile_from_name(db: Engine, profiles: list[ProteinProfile]):
    # Write the profiles to the DB
    with Session(db) as session:
        collection.add_profiles(session, profiles)

    for profile in profiles:
        # Build the profile from the DB
        with Session(db) as session:
            profile_from_db = collection.get_profile_from_name(session, profile.name)

        # They are the same
        _assert_profiles_equal(profile, profile_from_db)


def test_iter_all_profiles(db: Engine, profiles: list[ProteinProfile]):
    # Write the profiles to the DB
    with Session(db) as session:
        collection.add_profiles(session, profiles)

    with Session(db) as session:
        profile_iterator = iter_all_profiles(session)

    assert isinstance(profile_iterator, Iterator)

    # For each profile generated from the iterator, assert that it equates with the
    # corresponding profile. Note that the zip tests that order is maintained.
    for profile, profile_from_db in zip(profiles, profile_iterator, strict=True):
        _assert_profiles_equal(profile, profile_from_db)


def test_load_all_profiles(db: Engine, profiles: list[ProteinProfile]):
    # Write the profiles to the DB
    with Session(db) as session:
        collection.add_profiles(session, profiles)

    db_path = str(db.url).replace(DB_SCHEME, "")
    profiles_from_db = load_all_profiles(db_path)
    assert isinstance(profiles_from_db, list)

    # For each profile in the load_all_profiles result, assert that it equates with the
    # corresponding profile. Note that the zip tests that order is maintained.
    for profile, profile_from_db in zip(profiles, profiles_from_db, strict=True):
        _assert_profiles_equal(profile, profile_from_db)


@pytest.mark.parametrize("extension", RECOGNIZED_STRUCTURE_FILETYPES)
def test_export_structures(
    db: Engine, profiles: list[ProteinProfile], extension: str, cluster: LocalCluster
):
    with Session(db) as session:
        collection.add_profiles(session, profiles)

        # Store the exported structures, then load them from file
        with maybe_temp_directory() as tmpdir:
            with Client(cluster) as client:
                export_structures(session, tmpdir, extension=extension, client=client)

            loaded_structures = {
                structure_name_and_extension(path)[0]: load_structure(path)
                for path in tmpdir.glob(f"*{extension}")
            }

    # Now assert each profile structure matches the structure loaded from file
    for profile in profiles:
        assert structures_almost_equal(profile.structure, loaded_structures[profile.name])



from apb.application.foldseek.align import align_to_reference, create_reference_aligner
from apb.application.foldseek.base import (
    DEFAULT_CONVERTALIS_CONFIG,
    DEFAULT_SEARCH_CONFIG,
    convertalis,
    createdb,
    search,
)
from apb.application.foldseek.datatypes import FoldSeekConvertalisConfig, FoldSeekSearchConfig

__all__ = [
    "createdb",
    "search",
    "convertalis",
    "FoldSeekSearchConfig",
    "FoldSeekConvertalisConfig",
    "align_to_reference",
    "create_reference_aligner",
    "DEFAULT_CONVERTALIS_CONFIG",
    "DEFAULT_SEARCH_CONFIG",
]



"""Protein collection database management

This module provides functionality for adding protein profiles to a 'collection'
database and retrieving the contents. It handles the conversion of protein structures to
and from byte format for database storage and ensures the integrity of residue feature
data using pandera.
"""

from __future__ import annotations
import logging
from collections.abc import Generator, Iterable
from pathlib import Path
from typing import Any

import pandas as pd
from biotite.sequence.seqtypes import ProteinSequence
from dask.base import compute
from dask.delayed import Delayed, delayed
from dask.distributed import Client
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session
from sqlalchemy_utils import database_exists

import apb.database.collection.schema as collection_schema
from apb.database.utils import create_db, engine_from_path, engine_from_session, path_to_uri
from apb.profile import ProteinProfile
from apb.structure.convert import (
    save_structure,
    structure_from_bytes,
    structure_to_bytes,
)
from apb.structure.utils import RECOGNIZED_STRUCTURE_FILETYPES
from apb.types import Pathish

logging.getLogger("distributed").setLevel(logging.WARNING)
logging.getLogger("dask").setLevel(logging.WARNING)


PreprocessProfileReturnType = tuple[collection_schema.Protein, pd.DataFrame]


def load_all_profiles(path: Pathish) -> list[ProteinProfile]:
    """Loads and returns all protein profiles from a specified database.

    This function simplifies the process of retrieving protein profiles from a database
    by creating an internally managed SQLAlchemy session. By providing the path to the
    database, users can easily access all protein profiles without needing to manually
    handle session creation and management. This function is especially useful for users
    who frequently need to access protein profiles and want/need a straightforward
    method to do so.

    Args:
        path: The filesystem path or similar reference to the database.

    Returns:
        list[ProteinProfile]: A list of all protein profiles contained in the database.

    See Also:
        - iter_all_profiles: Function called internally to retrieve profiles.
    """
    with Session(get_engine(path, create=False)) as session:
        return list(iter_all_profiles(session))


def get_engine(path: Pathish, create: bool = False) -> Engine:
    """Load a SQLAlchemy engine for a protein collection

    Args:
        path:
            The file system path to the database. This should be convertible to a URI.
        create:
            Flag indicating whether to create the database if it does not exist.
            Defaults to `False`. If the path doesn't resolve to a DB and this is False,
            a ValueError is raised.

    Returns:
        Engine:
            A SQLAlchemy Engine instance connected to the specified database.

    Raises:
        ValueError: If the database does not exist and `create` is set to `False`.
    """
    uri = path_to_uri(path)
    engine = engine_from_path(path)

    if database_exists(uri):
        return engine

    if not create:
        raise ValueError(f"{path} doesn't exist")

    create_db(engine, collection_schema.Base, exist_ok=False)
    return engine


def add_profiles(session: Session, profiles: Iterable[ProteinProfile]) -> None:
    """Add profiles to the database

    Args:
        profiles:
            Iterable of protein profiles
    """
    add_preprocessed_profiles(session, (preprocess_profile(profile) for profile in profiles))


def add_preprocessed_profiles(
    session: Session, preprocessed_profiles: Iterable[PreprocessProfileReturnType]
) -> None:
    """Add preprocessed profiles to the database

    "Preprocessed profiles" are the return type of `preprocess_profile`.

    Args:
        profiles:
            An iterable of tuples where each tuple contains a Protein entry and a
            residue DataFrame.
    """
    residue_dfs: list[pd.DataFrame] = []

    for protein_entry, residue_df in preprocessed_profiles:
        session.add(protein_entry)
        session.flush()

        residue_df["protein_id"] = protein_entry.id
        residue_dfs.append(residue_df)

    if not len(residue_dfs):
        return

    # Validate the concatenated dataframe
    all_protein_residue_df = collection_schema.residue_dataframe_schema.validate(
        pd.concat(residue_dfs, ignore_index=True)
    )

    # Bulk insert the residue entries
    session.bulk_insert_mappings(
        collection_schema.Residue,  # type: ignore
        all_protein_residue_df.to_dict(orient="records"),
    )

    session.commit()


def preprocess_profile(profile: ProteinProfile) -> tuple[collection_schema.Protein, pd.DataFrame]:
    """Decompose a profile's data into table components prior to database insertion

    Returns:
        collection_schema.Protein:
            A collection_schema.Protein object.
        pd.DataFrame:
            The profile's residue features dataframe.
    """
    protein_entry = collection_schema.Protein(
        name=profile.name,
        sequence=str(profile.sequence),
        structure=structure_to_bytes(profile.structure),
        source=profile.source,
    )

    residue_df = profile.residue_features.reset_index()

    return protein_entry, residue_df


def _get_protein_by(session: Session, **kwargs: Any) -> collection_schema.Protein:
    """Helper function to retrieve a protein by given keyword arguments"""
    engine = engine_from_session(session)
    protein = session.query(collection_schema.Protein).filter_by(**kwargs).one_or_none()

    if protein is None:
        raise KeyError(f"Protein with {kwargs} is not present in {engine.url}.")

    return protein


def _build_protein_profile(session: Session, protein: collection_schema.Protein) -> ProteinProfile:
    """Helper function to build a ProteinProfile from a Protein SQLAlchemy object."""
    engine = engine_from_session(session)
    sequence = ProteinSequence(protein.sequence)
    structure = structure_from_bytes(protein.structure)  # type: ignore

    statement = session.query(collection_schema.Residue).filter_by(protein_id=protein.id).statement

    return ProteinProfile(
        name=str(protein.name),
        sequence=sequence,
        structure=structure,
        residue_features=(
            pd.read_sql(statement, engine)
            .pipe(collection_schema.residue_dataframe_schema.validate)
            .drop(["id", "protein_id", "residue_index"], axis=1)
        ),
        source=str(protein.source),
    )


def get_profile_from_index(session: Session, protein_id: int) -> ProteinProfile:
    protein = _get_protein_by(session, id=protein_id)
    return _build_protein_profile(session, protein)


def get_profile_from_name(session: Session, protein_name: str) -> ProteinProfile:
    protein = _get_protein_by(session, name=protein_name)
    return _build_protein_profile(session, protein)


def export_structures(
    session: Session,
    structure_dir: Pathish,
    client: Client,
    extension: str = ".pdb",
) -> Path:
    """Exports structures into a directory"""
    if extension not in RECOGNIZED_STRUCTURE_FILETYPES:
        raise ValueError(
            f"Unknown extension '{extension}'. Allowed extensions: {RECOGNIZED_STRUCTURE_FILETYPES}"
        )

    structure_dir = Path(structure_dir)
    if not structure_dir.exists():
        structure_dir.mkdir(parents=True)

    # Ideally, the input to a dask delayed chain is something simple like an integer or
    # the path to a file. This is because the object is a part of the DAG. Yet in this
    # case, the input to the structure is the protein SQLAlchemy object. This limits the
    # size we can make the task graph before running out of memory. I chose 5000 because
    # this creates a graph size <1Gb.
    batch_size = 5000

    save_tasks: list[Delayed] = []
    for protein in session.query(collection_schema.Protein).yield_per(batch_size):
        structure = delayed(structure_from_bytes)(protein.structure)
        save_task = delayed(save_structure)(structure_dir / f"{protein.name}{extension}", structure)
        save_tasks.append(save_task)

        if len(save_tasks) >= batch_size:
            compute(*save_tasks, scheduler=client)
            save_tasks = []

    if len(save_tasks):
        compute(*save_tasks, scheduler=client)

    return structure_dir


def iter_all_profiles(session: Session) -> Generator[ProteinProfile, None, None]:
    """Returns an iterator of all profiles in the DB

    This function iterates through all protein names in the database, and yields a
    protein profile for each one by querying the database based on the protein name. The
    function uses a generator, making it memory efficient for large datasets as it
    yields profiles one at a time.

    Examples:
        To load all profiles into memory, you can convert the generator to a list:

        >>> profiles = list(get_all_profiles(session))

        To iteratively print the length of each protein sequence:

        >>> for profile in get_all_profiles(session):
        >>>     print(f"{profile.name}: Sequence Length = {len(profile.sequence)}")

    See Also:
        - `load_all_profiles`
    """
    for name in get_all_protein_names(session):
        yield get_profile_from_name(session, name)


def get_all_protein_names(session: Session) -> list[str]:
    return [name for name, *_ in session.query(collection_schema.Protein.name).all()]



import re
from collections.abc import Iterable
from pathlib import Path

import pandas as pd
import pandera as pa

from apb.structure.utils import RECOGNIZED_STRUCTURE_FILETYPES
from apb.types import Pathish

# These are taken from `foldseek convertalis -h`. multimer measures (those prefixed with
# 'complex') are absent.
_alignment_dataframe_schema = pa.DataFrameSchema(
    {
        "query": pa.Column("str", required=False),
        "target": pa.Column("str", required=False),
        "evalue": pa.Column("float64", required=False),
        "gapopen": pa.Column("int64", required=False),
        "pident": pa.Column("float64", required=False),
        "fident": pa.Column("float64", required=False),
        "nident": pa.Column("int64", required=False),
        "qstart": pa.Column("int64", required=False),
        "qend": pa.Column("int64", required=False),
        "qlen": pa.Column("int64", required=False),
        "qaln": pa.Column("str", required=False),
        "tstart": pa.Column("int64", required=False),
        "tend": pa.Column("int64", required=False),
        "tlen": pa.Column("int64", required=False),
        "taln": pa.Column("str", required=False),
        "alnlen": pa.Column("int64", required=False),
        "bits": pa.Column("int64", required=False),
        "cigar": pa.Column("str", required=False),
        "qseq": pa.Column("str", required=False),
        "tseq": pa.Column("str", required=False),
        "qheader": pa.Column("str", required=False),
        "theader": pa.Column("str", required=False),
        "qalign": pa.Column("str", required=False),
        "talign": pa.Column("str", required=False),
        "mismatch": pa.Column("int64", required=False),
        "qcov": pa.Column("float64", required=False),
        "tcov": pa.Column("float64", required=False),
        "qset": pa.Column("str", required=False),
        "qsetid": pa.Column("int64", required=False),
        "tset": pa.Column("str", required=False),
        "tsetid": pa.Column("int64", required=False),
        "taxid": pa.Column("int64", required=False),
        "taxname": pa.Column("str", required=False),
        "taxlineage": pa.Column("str", required=False),
        "lddt": pa.Column("float64", required=False),
        "lddtfull": pa.Column("str", required=False),
        "qca": pa.Column("str", required=False),
        "tca": pa.Column("str", required=False),
        "t": pa.Column("str", required=False),
        "u": pa.Column("str", required=False),
        "qtmscore": pa.Column("float64", required=False),
        "ttmscore": pa.Column("float64", required=False),
        "alntmscore": pa.Column("float64", required=False),
        "rmsd": pa.Column("float64", required=False),
        "prob": pa.Column("float64", required=False),
    },
    coerce=True,
    strict=True,  # Enforces that all columns in the data exist in the schema
)


def format_alignment_tsv(frame: pd.DataFrame) -> pd.DataFrame:
    """Formats the alignment TSV DataFrame.

    Validates the DataFrame against a predefined schema and removes known file extensions
    from the query and target columns.

    Args:
        frame: The DataFrame containing alignment results.

    Returns:
        pd.DataFrame: The formatted dataframe.
    """
    frame = _alignment_dataframe_schema.validate(frame)

    # Foldseek only removes the first extension from file names. This is problematic
    # because if your file is `name.pdb.gz`, the name in the alignment table resolves to
    # `name.pdb` rather than `name`. We resolve this by compiling our accepted structure
    # formats into a regex.

    extension_pattern = r"|".join([re.escape(ext) for ext in RECOGNIZED_STRUCTURE_FILETYPES])

    if "query" in frame.columns:
        frame["query"] = frame["query"].str.replace(extension_pattern, "", regex=True)

    if "target" in frame.columns:
        frame["target"] = frame["target"].str.replace(extension_pattern, "", regex=True)

    return frame


def create_tsv_for_createdb(tsv_path: Pathish, structure_paths: Iterable[Pathish]) -> Path:
    """Create a TSV of structure paths that can be used as input from foldseek's `createdb`"""
    with open(tsv_path, "w") as fp:
        fp.writelines(f"{structure_path}\n" for structure_path in structure_paths)

    return Path(tsv_path)



import logging
import shutil
import subprocess
import tempfile
from collections.abc import Callable, Generator
from contextlib import contextmanager
from datetime import datetime, timedelta
from functools import wraps
from pathlib import Path
from typing import ParamSpec, TypeVar

import attrs
import cattr

from apb.types import Pathish
from apb.vendor.strenum import StrEnum

# Create a default converter for serialization/deserialization using cattrs (for details
# on cattrs, see https://catt.rs/en/stable/). If we expand more into designing
# dataclasses that are round-trip serializable, e.g. `FeatureRecipe` in
# apb/workflow/create_collection.py, we should create a serialize subpackage modelled
# after or identical to this:
# https://github.com/ekiefl/pooltool/tree/main/pooltool/serialize
converter = cattr.Converter()
converter.register_unstructure_hook(StrEnum, lambda v: v.value)


def is_executable(program: str) -> bool:
    """Check if a program name is accessible."""
    return shutil.which(program) is not None


P = ParamSpec("P")
T = TypeVar("T")


def require_dependency(program: str) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorator that checks if an external dependency is available before executing the function.

    Args:
        program: The name of the program dependency to check.

    Returns:
        A decorator function that will check for the dependency before executing the
        decorated function.

    Raises:
        RuntimeError: If the program is not available in PATH.
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            if not is_executable(program):
                raise RuntimeError(
                    f"Calling `{func.__name__}` requires external dependency '{program}', "
                    f"which is not found in PATH. Install {program} and ensure "
                    f"it's available in PATH."
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def run_command(command: list[str], log_stdout: bool = False) -> subprocess.CompletedProcess:
    """Runs a command using subprocess.

    Args:
        quiet:
            If True, stdout is not logged, otherwise it is. The stdout is not streamed,
            so no logging info until the subprocess completes.
    """

    if not len(command):
        raise ValueError("Empty command")

    if not is_executable(command[0]):
        command_string = " ".join(command)
        raise ValueError(f"'{command[0]}' from command '{command_string}' is not a program")

    result = subprocess.run(command, capture_output=True, text=True)

    if result.returncode != 0:
        raise RuntimeError(f"Command failed: {' '.join(command)}\n{result.stderr}")

    if log_stdout:
        logging.info(result.stdout)

    return result


@contextmanager
def maybe_temp_directory(
    existing_dir: Pathish | None = None,
) -> Generator[Path, None, None]:
    """Context manager that provides a temporary directory if an existing directory isn't passed.

    This context manager behaves like `tempfile.TemporaryDirectory` whenever
    `existing_dir` is None. Otherwise, it simply yields `existing_dir` after casting it
    as a pathlib.Path object.

    Args:
        existing_dir:
            An existing directory. If None (default), a new temporary
            directory is created.

    Yields:
        Path: The directory path.
    """
    if existing_dir is None:
        with tempfile.TemporaryDirectory() as tmpdir:
            yield Path(tmpdir)
    else:
        yield Path(existing_dir)


@attrs.define
class Timer:
    # Initialized
    msg: str = attrs.field(default="elapsed time")
    quiet: bool = attrs.field(default=False)
    suffix: bool = attrs.field(default=True)

    # Internal
    start: datetime = attrs.field(default=None, init=False)
    end: datetime = attrs.field(default=None, init=False)
    time: timedelta = attrs.field(default=None, init=False)

    def __enter__(self):
        self.start = datetime.now()
        return self

    def __exit__(self, *_):
        self.end = datetime.now()
        self.time = self.end - self.start

        if self.quiet:
            return

        if self.suffix:
            print(f"{self.time} {self.msg.strip()}")
        else:
            print(f"{self.msg.strip()} {self.time}")



from pathlib import Path

import pytest
from click.testing import CliRunner
from sqlalchemy.orm.session import Session

import apb.database.collection as collection
from apb._test_pointer import TEST_RAW_COLLECTION
from apb.workflow.create_collection import (
    FeatureCalculationConfig,
    cli,
    load_feature_config,
    save_feature_config,
)


@pytest.fixture
def runner():
    return CliRunner()


def test_help(runner: CliRunner):
    result = runner.invoke(cli, ["--help"])
    assert result.exit_code == 0
    assert "Create a protein collection database" in result.output


def test_create_feature_config_file(runner: CliRunner, tmp_path: Path):
    path = tmp_path / "config.json"

    # The CLI runs and creates the config
    result = runner.invoke(cli, ["config", str(path)])
    assert result.exit_code == 0
    assert path.exists()

    # The result can be structured into a FeatureRecipe object
    load_feature_config(path)


def test_create_config_file_exists(runner: CliRunner, tmp_path: Path):
    # Create the file
    path = tmp_path / "config.json"
    path.touch()

    # Run the CLI
    result = runner.invoke(cli, ["config", str(path)])

    # It fails because the file already exists
    assert result.exit_code != 0


def test_run_with_default_config(runner: CliRunner, tmp_path: Path):
    output_db = tmp_path / "output.db"

    # The CLI runs
    result = runner.invoke(
        cli,
        [
            "run",
            "-i",
            str(TEST_RAW_COLLECTION),
            "-o",
            str(output_db),
        ],
    )
    assert result.exit_code == 0

    # The DB can be read from
    with Session(collection.get_engine(output_db)) as session:
        collection.get_profile_from_index(session, 1)


def test_run_with_custom_config(runner: CliRunner, tmp_path: Path):
    output_db = tmp_path / "output.db"
    config_path = tmp_path / "config.json"

    # Create a custom config file
    save_feature_config(FeatureCalculationConfig.default(), config_path)

    # The CLI runs
    result = runner.invoke(
        cli,
        [
            "run",
            "-i",
            str(TEST_RAW_COLLECTION),
            "-o",
            str(output_db),
            "-c",
            str(config_path),
        ],
    )
    assert result.exit_code == 0

    # The DB can be read from
    with Session(collection.get_engine(output_db)) as session:
        collection.get_profile_from_index(session, 1)



from pathlib import Path

import pytest
from sqlalchemy import Column, Integer, String
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session, declarative_base
from sqlalchemy_utils import database_exists

from apb.database.utils import (
    create_db,
    engine_from_path,
    engine_from_session,
    path_to_uri,
)

# Define a test base and a simple model for testing
TestBase = declarative_base()


class User(TestBase):
    __tablename__ = "users"
    id = Column(Integer, primary_key=True)
    name = Column(String)


@pytest.fixture
def temp_db_path(tmp_path: Path):
    return tmp_path / "test.db"


@pytest.fixture
def engine(temp_db_path):
    """Create an SQLAlchemy engine connected to the temporary database."""
    return engine_from_path(temp_db_path)


def test_path_to_uri(temp_db_path):
    uri = path_to_uri(temp_db_path)
    assert uri.startswith("sqlite:///")
    assert uri.endswith(str(temp_db_path))


def test_engine_from_path(temp_db_path):
    engine = engine_from_path(temp_db_path)
    assert isinstance(engine, Engine)
    assert str(engine.url).startswith("sqlite:///")


def test_engine_from_session(engine):
    with Session(engine) as session:
        assert engine_from_session(session) == engine


def test_create_db(engine):
    assert not database_exists(engine.url)
    create_db(engine, TestBase)
    assert database_exists(engine.url)


def test_create_db_already_exists(engine):
    create_db(engine, TestBase)
    assert database_exists(engine.url)

    with pytest.raises(ValueError, match="Database already exists."):
        create_db(engine, TestBase, exist_ok=False)



from __future__ import annotations
from pathlib import Path

import attrs
import pytest
from biotite.sequence.seqtypes import ProteinSequence

from apb._test_pointer import TEST_STRUCTURES_ROOT, TEST_TMALIGN_ROOT
from apb.application.tmalign.core import TMAlignOutput, _parse_stdout, _run_cmd, run_tmalign


@attrs.define
class TMAlignRunBundle:
    """Houses the inputs, standard output, and structured output of a TMAlign run"""

    path1: Path
    path2: Path
    expected_standard_out: str
    expected_structured_output: TMAlignOutput

    @classmethod
    def create(
        cls, path1: Path, path2: Path, stdout_path: Path, output: TMAlignOutput
    ) -> TMAlignRunBundle:
        with open(stdout_path) as file:
            stdout = file.read()

        return TMAlignRunBundle(
            path1,
            path2,
            stdout,
            output,
        )


test_runs = [
    # Run 1
    TMAlignRunBundle.create(
        TEST_STRUCTURES_ROOT / "structure1.pdb",
        TEST_STRUCTURES_ROOT / "structure2.pdb",
        TEST_TMALIGN_ROOT / "tmalign.stdout1",
        TMAlignOutput(
            qseq_aligned="----PSIVARSNFNVCRLPGTPEALCATYTGCIIIPGATCPGDYA-",
            tseq_aligned="KSCCPTTAARNQYNICRLPGTPRPVC-------IISGTGCPP--RH",
            qtmscore=0.72678,
            ttmscore=0.79485,
            rmsd=0.58,
        ),
    ),
    # More runs can be added...
]


@pytest.fixture
def run() -> TMAlignRunBundle:
    return test_runs[0]


def test_filetype_integrity(run: TMAlignRunBundle):
    with pytest.raises(RuntimeError, match="Command failed"):
        run_tmalign(run.path1, Path("doesnt_exist.pdb"))

    with pytest.raises(RuntimeError, match="Command failed"):
        run_tmalign(Path("doesnt_exist.pdb"), run.path2)

    # Unallowed extensions fail.

    with pytest.raises(ValueError, match="TMAlign accepts the following extensions"):
        run_tmalign(run.path1.with_suffix(".cif"), run.path2)

    with pytest.raises(ValueError, match="TMAlign accepts the following extensions"):
        run_tmalign(run.path1, run.path2.with_suffix(".cif"))


@pytest.mark.parametrize("run", test_runs)
def test_run_cmd(run: TMAlignRunBundle):
    # stdout1 was generated from STRUCTURE1_PATH and STRUCTURE2_PATH, so we expect the
    # same output when we run _run_cmd.
    expected = run.expected_standard_out
    output = _run_cmd(run.path1, run.path2)

    # However, the chain names will differ, since they depend on which directory the
    # test is run within. For example e.g., "structure1.pdb" versus
    # "apb/align/tmalign/test_data/structure1.pdb". This means we cannot do a rote
    # comparison, and instead have to do a line-by-line comparison, skipping over the
    # chain name lines.
    expected_lines = expected.strip("\n").split("\n")
    output_lines = output.strip("\n").split("\n")
    for o_line, e_line in zip(output_lines, expected_lines, strict=True):
        if "Name of Chain" in o_line:
            continue
        assert o_line == e_line, f"Lines differ: {o_line} != {e_line}"


@pytest.mark.parametrize("run", test_runs)
def test_parse_stdout(run: TMAlignRunBundle):
    # Parsing the standard out should yield the structured out.
    assert _parse_stdout(run.expected_standard_out) == run.expected_structured_output


@pytest.mark.parametrize("run", test_runs)
def test_TMAlignOutput(run: TMAlignRunBundle):
    # Strip the gaps of the aligned tseq.
    expected_tseq = run.expected_structured_output.tseq_aligned.replace("-", "")

    # Ensure cached property `tseq` matches.
    assert run.expected_structured_output.tseq == expected_tseq


@pytest.mark.parametrize("run", test_runs)
def test_inverse_relationship(run: TMAlignRunBundle):
    """Test the inverse relationship

    TMalign has an inverse relationship with its inputs, whereby the alignment of query
    to target can be derived from the inverse alignment (target to query). While this is
    conceptually true, and in practice it is almost always true, sometimes TMAlign has
    different outputs. However, these differences are rare and slight.
    """
    out1 = run_tmalign(run.path1, run.path2)
    out2 = run_tmalign(run.path2, run.path1)

    assert out1.qseq_aligned == out2.tseq_aligned
    assert out1.tseq_aligned == out2.qseq_aligned
    assert out1.qtmscore == out2.ttmscore
    assert out1.ttmscore == out2.qtmscore


@pytest.mark.parametrize("run", test_runs)
def test_to_biotite(run: TMAlignRunBundle):
    out = run_tmalign(run.path1, run.path2)
    alignment = out.to_biotite()

    assert alignment.sequences[0] == ProteinSequence(out.qseq)
    assert alignment.sequences[1] == ProteinSequence(out.tseq)

    # The entirety of both sequences are found in the trace. Test this by observing the
    # ends of each sequence (0 and len(seq)-1).
    assert 0 in alignment.trace[:, 0]
    assert 0 in alignment.trace[:, 1]
    assert len(out.qseq) - 1 in alignment.trace[:, 0]
    assert len(out.tseq) - 1 in alignment.trace[:, 1]



"""Fold sequences using Arcadia's folding server.

Usage:
    $ apb-fold --help
"""

from __future__ import annotations
import asyncio
import logging
import random
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import Any

import aiohttp
import attrs
import biotite.sequence.io.fasta as fasta
import cattrs
import pandas as pd
import typer
import yaml
from dateutil.parser import isoparse
from rich.console import Console
from rich.logging import RichHandler
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    Progress,
    ProgressColumn,
    TaskID,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rich.prompt import Prompt
from slugify import slugify

from apb.config import CONFIG_DIR, ensure_config_dir_exists

app = typer.Typer(
    help="Fold protein sequences using Arcadia's folding server.",
    add_completion=False,
    pretty_exceptions_show_locals=False,
)
console = Console()

logging.basicConfig(level="INFO", format="%(message)s", handlers=[RichHandler(console=console)])
logger = logging.getLogger("rich")

# Create a converter and register a structure hook for datetime
converter = cattrs.Converter()
converter.register_structure_hook(datetime, lambda dt_str, _: isoparse(dt_str) if dt_str else None)

FOLD_CONFIG_DIR = CONFIG_DIR / "fold_server"
FOLD_CREDENTIALS = FOLD_CONFIG_DIR / "credentials.yaml"

FETCH_TIMEOUT: int = 60 * 60 * 10  # Total time to keep retrying (in seconds)
INITIAL_RETRY_DELAY: float = 10  # Initial delay in seconds
BACKOFF_FACTOR: float = 1.03  # Exponential backoff factor


def get_url_and_token() -> tuple[str, str]:
    ensure_config_dir_exists()

    if not FOLD_CREDENTIALS.exists():
        console.print(
            Panel(
                "[red]Credentials not found![/red]\n\n"
                "Please run [bold]apb-fold setup[/bold] first to configure your API credentials.",
                title="Configuration Required",
                border_style="red",
            )
        )
        raise typer.Exit(code=1)

    with open(FOLD_CREDENTIALS) as f:
        config = yaml.safe_load(f)

    api_url = config.get("api_url")
    bearer_token = config.get("bearer_token")

    if not api_url or not bearer_token:
        console.print(
            Panel(
                f"[red]api_url and/or bearer_token missing[/red]\n\n"
                f"Run [bold]apb-fold setup[/bold] or update {FOLD_CREDENTIALS} manually.",
                title="Configuration Error",
                border_style="red",
            )
        )
        raise typer.Exit(code=1)

    return api_url, bearer_token


@attrs.define
class Response:
    # Present in /fold and /result responses
    job_id: str
    sequence: str
    pdb_uri: str | None
    status: str
    device_type: str
    failure_reason: str | None
    created_at: datetime
    updated_at: datetime
    folding_model: str

    # Present in /result resonses
    pdb_content: str | None = None

    @property
    def runtime(self) -> float:
        """Number of sequences between updated_at and created_at"""
        return (self.updated_at - self.created_at).total_seconds()

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Response:
        return converter.structure(data, Response)


async def submit_job(
    session: aiohttp.ClientSession, sequence: str, api_url: str, headers: dict[str, str]
) -> Response | None:
    url: str = f"{api_url}/fold"
    data: dict[str, str] = {"protein_sequence": sequence}

    async with session.post(url, json=data, headers=headers) as response:
        if response.status == 200:
            response_json = await response.json()
            return Response.from_dict(response_json)

    console.print(
        f"[bold]Failed[/bold] submission for '{sequence}'. Response: {response.status}", style="red"
    )

    return None


def _get_most_recently_succeeded(results: list[Response]) -> Response | None:
    succeeded_results = [result for result in results if result.status == "SUCCEEDED"]
    if succeeded_results:
        return max(succeeded_results, key=lambda r: r.updated_at)

    return None


async def fetch_result(
    session: aiohttp.ClientSession, job_id: str | None, api_url: str, headers: dict[str, str]
) -> Response | None:
    if job_id is None:
        return None

    url: str = f"{api_url}/result/{job_id}"

    attempt: int = 0
    delay: float = INITIAL_RETRY_DELAY
    cumulative_delay: float = 0

    while cumulative_delay + delay < FETCH_TIMEOUT:
        async with session.get(url, headers=headers) as response:
            if response.status == 200:
                response_json = await response.json()

                # /result endpoint always returns a list of jobs
                results = [Response.from_dict(job_json) for job_json in response_json]
                result = _get_most_recently_succeeded(results)

                if result is not None:
                    msg = (
                        f"[bold]Folded[/bold] {result.job_id} "
                        f"({result.updated_at}) "
                        f"[bold][magenta]{result.sequence}"
                    )
                    console.print(msg, style="green")
                    return result
            else:
                console.print(
                    f"Error fetching job {job_id}. Status code: {response.status}. "
                    f"Attempt {attempt + 1}.",
                    style="red",
                )

        logger.debug(f"Job {job_id}: attempt {attempt} failed. Trying again in {delay} seconds.")
        jitter = random.uniform(0, delay * 0.1)
        await asyncio.sleep(delay + jitter)

        cumulative_delay += delay
        attempt += 1
        delay = INITIAL_RETRY_DELAY * (BACKOFF_FACTOR**attempt)

    console.print(
        f"Failed to fetch result for job {job_id} after exceeding the timeout of "
        f"{FETCH_TIMEOUT} seconds.",
        style="red",
    )

    return None


def _update_progress(task: asyncio.Task, progress: Progress, progress_task_id: TaskID) -> None:
    assert task.done()
    progress.advance(progress_task_id)


async def fold(sequences: dict[str, str], output: Path | str, quiet: bool = False) -> Path:
    console.quiet = quiet

    api_url, token = get_url_and_token()
    headers = {"Content-Type": "application/json", "Authorization": f"Bearer {token}"}

    output = Path(output)
    output.mkdir(exist_ok=True)

    progress_columns: list[ProgressColumn] = [
        TextColumn("{task.description}"),
        BarColumn(),
        TaskProgressColumn(),
        TextColumn("[progress.percentage]{task.completed}/{task.total}"),
        TimeElapsedColumn(),
        TimeRemainingColumn(),
    ]

    if not quiet:
        msg = (
            "[green][bold] Job Information[/bold]\n"
            "• Once all jobs have submitted, they will continue running on the server even\n"
            "  if you close this terminal\n"
            "• At that point you can safely cancel this command or close your laptop - your\n"
            "  jobs will keep processing\n"
            "• To retrieve results later, simply run this same command again - previously\n"
            "  completed jobs are cached and won't be re-folded\n"
            "• Once all jobs are finished, this command will download the results to your local"
            "[/green]"
        )
        console.print(
            Panel(
                msg,
                border_style="green",
                title="[bold]Fold Server Jobs[/bold]",
                expand=False,
            )
        )

    with Progress(*progress_columns, console=console) as progress:
        submitting_progress_id = progress.add_task("[cyan]Submitting...", total=len(sequences))
        fetching_progress_id = progress.add_task("[cyan]Fetching...", total=len(sequences))

        update_submit = partial(
            _update_progress, progress=progress, progress_task_id=submitting_progress_id
        )
        update_fetch = partial(
            _update_progress, progress=progress, progress_task_id=fetching_progress_id
        )

        async with aiohttp.ClientSession() as session:
            submission_tasks: list[asyncio.Task] = []
            for sequence in sequences.values():
                task = asyncio.create_task(submit_job(session, sequence, api_url, headers))
                task.add_done_callback(update_submit)
                submission_tasks.append(task)

            submission_responses: list[Response | None] = await asyncio.gather(*submission_tasks)

            fetch_tasks: list[asyncio.Task] = []
            for job in submission_responses:
                task = asyncio.create_task(
                    fetch_result(session, job.job_id if job else None, api_url, headers)
                )
                task.add_done_callback(update_fetch)
                fetch_tasks.append(task)

            fold_results: list[Response | None] = await asyncio.gather(*fetch_tasks)

        progress.refresh()

    failures: list[dict[str, str | None]] = []

    for (seq_id, seq), fold_request, fold_result in zip(
        sequences.items(), submission_responses, fold_results, strict=True
    ):
        if fold_request is None:
            # Failed request
            failures.append(
                {"seq_id": seq_id, "failure": "submission", "job_id": None, "sequence": seq}
            )
        elif fold_result is None:
            # Failed fold / timeout
            failures.append(
                {
                    "seq_id": seq_id,
                    "failure": "fetch",
                    "job_id": fold_request.job_id,
                    "sequence": seq,
                }
            )
        else:
            # Success
            assert fold_result.status == "SUCCEEDED"
            assert fold_result.pdb_content is not None
            filename_safe_seq_id = slugify(seq_id)
            output_path = output / f"{filename_safe_seq_id}.pdb"
            with open(output_path, "w") as fp:
                fp.write(fold_result.pdb_content)

    if failures:
        failure_df = pd.DataFrame(failures)
        failure_path = output / "_failures.tsv"
        failure_df.to_csv(failure_path, sep="\t", index=False)

    console.print(f"Results saved in [bold]{output}[/bold]", style="green")

    return output


def setup_credentials() -> None:
    """Setup API credentials for the fold server."""
    ensure_config_dir_exists()

    console.print(
        Panel(
            "[bold cyan]Arcadia's Folding Server Setup[/bold cyan]\n\n"
            "This will configure your API credentials for Arcadia's folding server.\n"
            "You'll need to provide your API URL and Bearer Token. Please ping the \n"
            "Software Team for these two pieces of information.",
            title="Welcome",
            border_style="cyan",
        )
    )

    FOLD_CONFIG_DIR.mkdir(parents=True, exist_ok=True)

    api_url = Prompt.ask("[bold]API URL[/bold]").rstrip("/")
    bearer_token = Prompt.ask("[bold]Bearer Token[/bold]", password=True)

    if not api_url or not bearer_token:
        console.print("[red]Both API URL and Bearer Token are required![/red]")
        raise typer.Exit(code=1)

    credentials = {"api_url": api_url, "bearer_token": bearer_token}

    with open(FOLD_CREDENTIALS, "w") as f:
        yaml.safe_dump(credentials, f, default_flow_style=False)

    console.print(
        Panel(
            f"[green]✓ Credentials saved successfully![/green]\n\n"
            f"Configuration saved to: [dim]{FOLD_CREDENTIALS}[/dim]\n\n"
            "You can now use [bold]apb-fold run[/bold] to submit folding jobs.",
            title="Setup Complete",
            border_style="green",
        )
    )


@app.command()
def setup() -> None:
    """Configure API credentials for the fold server."""
    setup_credentials()


@app.command()
def run(
    fasta_file: Path = typer.Option(  # noqa: B008
        ...,
        "--fasta-file",
        "-f",
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        help="Path to FASTA file containing sequences to fold.",
    ),
    output_dir: Path = typer.Option(  # noqa: B008
        ...,
        "--output-dir",
        "-o",
        file_okay=False,
        dir_okay=True,
        writable=True,
        help="Directory where folded structures (.pdb) will be written.",
    ),
    quiet: bool = typer.Option(  # noqa: B008
        False,
        "--quiet",
        "-q",
        help="Suppress progress output.",
    ),
) -> None:
    """Fold protein sequences from a FASTA file."""
    sequences = dict(fasta.FastaFile.read(fasta_file).items())
    asyncio.run(fold(sequences, output_dir, quiet=quiet))


if __name__ == "__main__":
    app()



import numpy as np
import pandas as pd
from biotite.structure.residues import apply_residue_wise, get_residues
from biotite.structure.sasa import sasa

from apb.structure.features.accessibility.datatypes import MaxSASASource, SolventAccessibilityConfig
from apb.structure.features.accessibility.normalization import get_max_sasa_map
from apb.types import ProteinStructure

DEFAULT_CONFIG = SolventAccessibilityConfig.default()


def calc_solvent_accessibility(
    structure: ProteinStructure,
    config: SolventAccessibilityConfig = DEFAULT_CONFIG,
) -> pd.Series:
    atomic_sasa = sasa(structure, vdw_radii="Single", point_number=config.point_number)
    solv_acc = apply_residue_wise(structure, atomic_sasa, np.sum)

    assert solv_acc is not None

    if config.normalize:
        max_sasa_map = get_max_sasa_map(config.source)
        _, residues = get_residues(structure)

        for idx, residue in enumerate(residues):
            solv_acc[idx] /= max_sasa_map[residue]

        # Estimated RSAs can exceed slightly > 1. We enforce these residues to have RSA := 1
        solv_acc[solv_acc > 1.0] = 1.0

    return pd.Series(
        solv_acc,
        dtype="float64",
        name=config.feature_name,
    )


__all__ = [
    "SolventAccessibilityConfig",
    "MaxSASASource",
    "calc_solvent_accessibility",
]



from pathlib import Path

import attrs
import pandas as pd
import pytest

from apb._test_pointer import TEST_RAW_COLLECTION
from apb.application.foldseek.base import (
    DEFAULT_CONVERTALIS_CONFIG,
    DEFAULT_SEARCH_CONFIG,
    convertalis,
    createdb,
    format_alignment_tsv,
    search,
)
from apb.application.foldseek.datatypes import FoldSeekConvertalisConfig, FoldSeekSearchConfig
from apb.utils import maybe_temp_directory


def test_createdb(tmp_path: Path):
    input_path = TEST_RAW_COLLECTION
    output_path_prefix = tmp_path / "output_db"

    # Sensible error message when path doesn't exist.
    with pytest.raises(RuntimeError, match="does not exist"):
        createdb(Path("nonexistant_path"), output_path_prefix)

    # Sensible error message when output requested is a directory.
    (tmp_path / "directory").mkdir()
    with pytest.raises(RuntimeError, match="Can not open"):
        createdb(input_path, (tmp_path / "directory"))

    # This runs without error and all expected DB files exist.
    createdb(input_path, output_path_prefix)
    assert output_path_prefix.exists()
    assert output_path_prefix.with_suffix(".dbtype").exists()
    assert output_path_prefix.with_suffix(".index").exists()
    assert output_path_prefix.with_suffix(".lookup").exists()
    assert output_path_prefix.with_suffix(".source").exists()

    output_ss = output_path_prefix.parent / (output_path_prefix.name + "_ss")
    assert output_ss.exists()
    assert output_ss.with_suffix(".dbtype").exists()
    assert output_ss.with_suffix(".index").exists()

    output_ca = output_path_prefix.parent / (output_path_prefix.name + "_ca")
    assert output_ca.exists()
    assert output_ca.with_suffix(".dbtype").exists()
    assert output_ca.with_suffix(".index").exists()

    output_h = output_path_prefix.parent / (output_path_prefix.name + "_h")
    assert output_h.exists()
    assert output_h.with_suffix(".dbtype").exists()
    assert output_h.with_suffix(".index").exists()


def test_search_config_parameters():
    default_config = FoldSeekSearchConfig()

    # Default parmeters are empty.
    assert default_config.parameters() == []

    # Each parameter modifies the parameters accordingly.

    config = attrs.evolve(default_config, backtrace=True)
    assert config.parameters() == ["-a"]

    config = attrs.evolve(default_config, backtrace=False)
    assert config.parameters() == []

    config = attrs.evolve(default_config, alignment_type=0)
    assert config.parameters() == ["--alignment-type", "0"]

    config = attrs.evolve(default_config, prefilter_mode=1)
    assert config.parameters() == ["--prefilter-mode", "1"]

    config = attrs.evolve(default_config, exact_tmscore=1)
    assert config.parameters() == ["--exact-tmscore", "1"]

    config = attrs.evolve(default_config, tmalign_fast=0)
    assert config.parameters() == ["--tmalign-fast", "0"]


def test_convertalis_config_parameters():
    default_config = FoldSeekConvertalisConfig()

    # Default parmeters are empty.
    assert default_config.parameters() == []

    # Each parameter modifies the parameters accordingly.

    config = attrs.evolve(default_config, exact_tmscore=1)
    assert config.parameters() == ["--exact-tmscore", "1"]

    config = attrs.evolve(default_config, format_output="query,qlen,target,tlen")
    assert config.parameters() == ["--format-output", "query,qlen,target,tlen"]


def test_search(tmp_path: Path):
    query_path = tmp_path / "query_db"
    target_path = tmp_path / "target_db"
    alignment_path_prefix = tmp_path / "ali"
    config = FoldSeekSearchConfig()

    # The DB paths must exist.
    with pytest.raises(RuntimeError, match="does not exist"):
        search(query_path, target_path, alignment_path_prefix, config)

    createdb(TEST_RAW_COLLECTION, query_path)
    createdb(TEST_RAW_COLLECTION, target_path)

    # This runs without error and all expected DB files exist.
    search(query_path, target_path, alignment_path_prefix, config)
    assert alignment_path_prefix.with_suffix(".dbtype").exists()
    assert alignment_path_prefix.with_suffix(".index").exists()


@pytest.fixture
def convertalis_missing_db_paths(tmp_path: Path) -> tuple[Path, Path, Path]:
    """Fixture of the non-existant paths required for convertalis"""
    query_path = tmp_path / "query_db"
    target_path = tmp_path / "target_db"
    alignment_path_prefix = tmp_path / "ali"
    return query_path, target_path, alignment_path_prefix


@pytest.fixture
def convertalis_db_paths(
    convertalis_missing_db_paths: tuple[Path, Path, Path],
) -> tuple[Path, Path, Path]:
    """Fixture of the existant paths required for convertalis"""
    query_path, target_path, alignment_path_prefix = convertalis_missing_db_paths

    # Create the query, target, and alignment DBs
    createdb(TEST_RAW_COLLECTION, query_path)
    createdb(TEST_RAW_COLLECTION, target_path)
    search(query_path, target_path, alignment_path_prefix)

    return query_path, target_path, alignment_path_prefix


def test_convertalis_db_paths_dont_exist(convertalis_missing_db_paths: tuple[Path, Path, Path]):
    query_path, target_path, alignment_path_prefix = convertalis_missing_db_paths

    # The query and target DBs must exist
    with pytest.raises(RuntimeError, match="does not exist"):
        convertalis(query_path, target_path, alignment_path_prefix)

    # Both DBs must exist
    createdb(TEST_RAW_COLLECTION, query_path)
    with pytest.raises(RuntimeError, match="does not exist"):
        convertalis(query_path, target_path, alignment_path_prefix)


def test_convertalis_alignment_doesnt_exist(convertalis_missing_db_paths: tuple[Path, Path, Path]):
    query_path, target_path, alignment_path_prefix = convertalis_missing_db_paths

    # Create the query and target DBs
    createdb(TEST_RAW_COLLECTION, query_path)
    createdb(TEST_RAW_COLLECTION, target_path)

    # The alignment must exist.
    with pytest.raises(RuntimeError, match="does not exist"):
        convertalis(query_path, target_path, alignment_path_prefix)


def test_convertalis_runs_successfully(convertalis_db_paths: tuple[Path, Path, Path]):
    query_path, target_path, alignment_path_prefix = convertalis_db_paths

    with maybe_temp_directory() as tmpdir:
        tsv_output = tmpdir / "out.tsv"

        # With all files in place, convertalis runs successfully.
        result_df = convertalis(query_path, target_path, alignment_path_prefix, tsv_output)

        # Verify that the expected TSV file exists
        assert tsv_output.exists()

        # The raw TSV has the same column headers as the output result.
        tsv_as_frame = pd.read_csv(tsv_output, sep="\t")
        assert tsv_as_frame.columns.equals(result_df.columns)

        # But they aren't equal until the TSV is validated by pandera.
        assert not tsv_as_frame.equals(result_df)
        assert format_alignment_tsv(tsv_as_frame).equals(result_df)


def test_convertalis_custom_format(convertalis_db_paths: tuple[Path, Path, Path]):
    query_path, target_path, alignment_path_prefix = convertalis_db_paths
    config = FoldSeekSearchConfig()

    # Create the query, target, and alignment DBs
    createdb(TEST_RAW_COLLECTION, query_path)
    createdb(TEST_RAW_COLLECTION, target_path)
    search(query_path, target_path, alignment_path_prefix, config)

    # Custom formats are possible.
    custom_config = attrs.evolve(DEFAULT_CONVERTALIS_CONFIG, format_output="query,qlen,qseq,pident")
    custom_result = convertalis(
        query_path,
        target_path,
        alignment_path_prefix,
        config=custom_config,
    )
    assert custom_config.format_output is not None
    assert custom_result.columns.tolist() == custom_config.format_output.split(",")


def test_convertalis_unrecognized_format(convertalis_db_paths: tuple[Path, Path, Path]):
    query_path, target_path, alignment_path_prefix = convertalis_db_paths
    config = FoldSeekSearchConfig()

    # Create the query, target, and alignment DBs
    createdb(TEST_RAW_COLLECTION, query_path)
    createdb(TEST_RAW_COLLECTION, target_path)
    search(query_path, target_path, alignment_path_prefix, config)

    # Unrecognized metrics are caught.
    custom_config = attrs.evolve(DEFAULT_CONVERTALIS_CONFIG, format_output="query,target,foo")
    with pytest.raises(RuntimeError, match="Format code foo does not exist"):
        convertalis(
            query_path,
            target_path,
            alignment_path_prefix,
            config=custom_config,
        )


def test_convertalis_with_backtrace(convertalis_db_paths: tuple[Path, Path, Path]):
    query_path, target_path, alignment_path_prefix = convertalis_db_paths
    config = FoldSeekSearchConfig()

    # Create the query, target, and alignment DBs
    createdb(TEST_RAW_COLLECTION, query_path)
    createdb(TEST_RAW_COLLECTION, target_path)
    search(query_path, target_path, alignment_path_prefix, config)

    # Some metrics require `search` to have been run with backtrace!
    custom_config = attrs.evolve(DEFAULT_CONVERTALIS_CONFIG, format_output="query,target,qaln,taln")
    with pytest.raises(RuntimeError, match="Please recompute the alignment with the -a flag"):
        convertalis(
            query_path,
            target_path,
            alignment_path_prefix,
            config=custom_config,
        )

    # After recomputing the search with the backtrace, it works.
    search(
        query_path,
        target_path,
        alignment_path_prefix,
        attrs.evolve(DEFAULT_SEARCH_CONFIG, backtrace=True),
    )
    custom_result = convertalis(
        query_path,
        target_path,
        alignment_path_prefix,
        config=custom_config,
    )
    assert custom_config.format_output is not None
    assert custom_result.columns.tolist() == custom_config.format_output.split(",")



import pandas as pd
import pytest
from biotite.sequence.seqtypes import ProteinSequence
from biotite.structure.residues import get_residues

from apb._test_pointer import structure_paths
from apb.structure.convert import load_structure
from apb.structure.features.secondary_structure import calc_3_state_secondary_structure, ss3_dtype

STRUCTURE_PATHS = structure_paths()


@pytest.mark.parametrize("path", STRUCTURE_PATHS)
def test_calc_ss_output(path):
    structure = load_structure(path)
    sequence = ProteinSequence(get_residues(structure)[1])
    out = calc_3_state_secondary_structure(structure)

    # Dtype is right
    assert out.dtype == ss3_dtype, "Values should be categorical dtype"

    # Name is correct
    assert out.name == "secondary_structure"

    # Length matches sequence
    assert len(sequence) == len(out)

    # There are no null values
    assert not out.isnull().any()

    # Assert indices are properly ordered and start at 0
    assert out.index.equals(pd.Index(range(len(sequence))))



import pandas as pd
from biotite.structure.sse import annotate_sse

from apb.types import ProteinStructure, ss3_dtype

FEATURE_NAME = "secondary_structure"


def calc_3_state_secondary_structure(structure: ProteinStructure) -> pd.Series:
    sse_array = annotate_sse(structure)
    return pd.Series(sse_array, dtype=ss3_dtype, name=FEATURE_NAME)


__all__ = [
    "calc_3_state_secondary_structure",
]



import tempfile
from pathlib import Path

import pytest

from apb.utils import is_executable, maybe_temp_directory, require_dependency, run_command


def test_is_executable():
    # These are programs
    assert is_executable("ln")
    assert is_executable("python")

    # These are not
    assert not is_executable("not_a_program")
    assert not is_executable("definitely not a program")
    assert not is_executable("")


def test_require_dependency():
    @require_dependency("python")
    def this_runs():
        pass

    this_runs()

    @require_dependency("not_a_program")
    def this_doesnt_run():
        pass

    with pytest.raises(RuntimeError, match="not found in PATH"):
        this_doesnt_run()

    # Now test that decorator stacking behavior works.

    @require_dependency("ln")
    @require_dependency("python")
    def this_also_runs():
        pass

    this_also_runs()

    @require_dependency("python")
    @require_dependency("not_a_program")
    def this_also_doesnt_run():
        pass

    with pytest.raises(RuntimeError, match="not found in PATH"):
        this_also_doesnt_run()


def test_run_command():
    with pytest.raises(ValueError, match="Empty command"):
        run_command([])

    with pytest.raises(ValueError, match="not a program"):
        run_command(["not_a_command", "foo"])

    with pytest.raises(RuntimeError, match="name 'some_variable' is not defined"):
        run_command(["python", "-c", "print(some_variable)"])

    assert run_command(["echo", "hi"]).stdout.strip() == "hi"
    assert run_command(["python", "-c", "print(4*4)"]).stdout.strip() == "16"


def test_temp_directory_with_no_tmpdir():
    """Test the temp directory context manager without a provided temporary directory."""
    with maybe_temp_directory() as tmpdir:
        assert tmpdir.exists(), "Temporary directory should exist"
        temp_file = tmpdir / "example.txt"
        temp_file.write_text("Hello, World!")
        assert temp_file.read_text() == "Hello, World!"
    assert not tmpdir.exists(), "Temporary directory should be deleted after exiting the context"


def test_temp_directory_with_existing_tmpdir():
    """Test the temp directory context manager with a provided existing temporary directory."""
    with tempfile.TemporaryDirectory() as existing_tmpdir:
        existing_path = Path(existing_tmpdir)

        with maybe_temp_directory(existing_path) as tmpdir:
            assert tmpdir == existing_path, "Should use the existing directory"

            temp_file = tmpdir / "example.txt"
            temp_file.write_text("Temporary check")

            assert temp_file.read_text() == "Temporary check"

        assert tmpdir.exists(), "Existing temp dir should not be deleted after exiting the context"
        assert (tmpdir / "example.txt").exists(), "File should exist in the provided directory"



from pathlib import Path

import attrs


@attrs.define(frozen=True)
class FoldseekOutputFileNotFoundError(Exception):
    path: Path
    command: list[str]

    @property
    def command_string(self) -> str:
        return " ".join(self.command)

    def __str__(self) -> str:
        return (
            f"Expected foldseek (`{self.command_string}`) to produce file '{self.path}', "
            "but it doesn't exist."
        )


@attrs.define(frozen=True)
class FoldSeekSearchConfig:
    """Parameter structuring for foldseek's search subcommand

    See `foldseek search -h` for details.
    """

    backtrace: bool | None = attrs.field(default=None)
    prefilter_mode: int | None = attrs.field(default=None)
    alignment_type: int | None = attrs.field(default=None)
    tmalign_fast: int | None = attrs.field(default=None)
    exact_tmscore: int | None = attrs.field(default=None)

    def parameters(self) -> list[str]:
        flag_strings = []

        if self.backtrace:
            flag_strings.append("-a")

        if self.prefilter_mode is not None:
            flag_strings.extend(["--prefilter-mode", str(self.prefilter_mode)])

        if self.alignment_type is not None:
            flag_strings.extend(["--alignment-type", str(self.alignment_type)])

        if self.tmalign_fast is not None:
            flag_strings.extend(["--tmalign-fast", str(int(self.tmalign_fast))])

        if self.exact_tmscore is not None:
            flag_strings.extend(["--exact-tmscore", str(int(self.exact_tmscore))])

        return flag_strings


@attrs.define(frozen=True)
class FoldSeekConvertalisConfig:
    """Parameter structuring for foldseek's convertalis subcommand

    See `foldseek convertalis -h` for details.
    """

    format_output: str | None = attrs.field(default=None)
    exact_tmscore: int | None = attrs.field(default=None)

    def parameters(self) -> list[str]:
        flag_strings = []

        if self.format_output is not None:
            flag_strings.extend(["--format-output", self.format_output])

        if self.exact_tmscore is not None:
            flag_strings.extend(["--exact-tmscore", str(int(self.exact_tmscore))])

        return flag_strings



from apb.application.tmalign.align import (
    TMAlignOutput,
    align_many_to_many,
    align_to_reference,
    load_alignments_json,
    save_alignments_json,
)
from apb.application.tmalign.core import run_tmalign

__all__ = [
    "run_tmalign",
    "align_many_to_many",
    "align_to_reference",
    "load_alignments_json",
    "save_alignments_json",
    "TMAlignOutput",
]



from pathlib import Path

import pytest
from biotite.structure.residues import get_residues
from dask.distributed import Client, LocalCluster

from apb._test_pointer import TEST_RAW_COLLECTION
from apb.application.tmalign.align import (
    align_many_to_many,
    align_to_reference,
    load_alignments_json,
    save_alignments_json,
)
from apb.structure.convert import load_structure, save_structure
from apb.structure.utils import structure_name_and_extension
from apb.types import not_none

QUERY_STRUCTURE_COUNT = 3
TARGET_STRUCTURE_COUNT = 4


@pytest.fixture
def target_structures(tmpdir: Path) -> dict[str, Path]:
    output: dict[str, Path] = {}

    for count, path in enumerate(TEST_RAW_COLLECTION.glob("*")):
        if count == TARGET_STRUCTURE_COUNT:
            break

        name, ext = structure_name_and_extension(path)

        if ext == ".pdb":
            pdb_path = path
        else:
            pdb_path = tmpdir / f"{name}.pdb"
            save_structure(pdb_path, load_structure(path))

        output[not_none(name)] = pdb_path

    return output


@pytest.fixture
def query_structures(target_structures: dict[str, Path]) -> dict[str, Path]:
    output: dict[str, Path] = {}

    for count, (target, structure) in enumerate(target_structures.items()):
        if count == QUERY_STRUCTURE_COUNT:
            break

        output[target] = structure

    return output


@pytest.fixture
def reference_structure(target_structures: dict[str, Path]) -> Path:
    return list(target_structures.values())[0]


@pytest.fixture
def structure_path_with_bad_extension(reference_structure: Path) -> Path:
    return Path(reference_structure).with_suffix(".pdb.gz")


def test_align_many_to_many(
    query_structures: dict[str, Path],
    target_structures: dict[str, Path],
    tmpdir: Path,
    cluster: LocalCluster,
):
    with Client(cluster) as client:
        output = align_many_to_many(query_structures, target_structures, client)

    expected_length = len(query_structures) * len(target_structures)

    length = 0
    for query in output:
        length += len(output[query])

        for target in output[query]:
            # Every entry is populated.
            assert output[query][target] is not None

            # Assert inverse relationship where applicable
            if target in output and query in output[target]:
                alignment = output[query][target]
                inverse = output[target][query]
                assert alignment.qseq_aligned == inverse.tseq_aligned
                assert alignment.tseq_aligned == inverse.qseq_aligned
                assert alignment.ttmscore == inverse.qtmscore
                assert alignment.qtmscore == inverse.ttmscore
                assert alignment.aligned == inverse.aligned
                assert alignment.rmsd == inverse.rmsd

    assert length == expected_length

    # Test round trip serialization
    assert output == load_alignments_json(save_alignments_json(tmpdir / "data.json", output))


def test_align_to_reference(
    reference_structure: Path, target_structures: dict[str, Path], cluster: LocalCluster
):
    with Client(cluster) as client:
        result = align_to_reference(reference_structure, target_structures, client)

    # The index length matches the reference length
    ref_length = len(get_residues(load_structure(reference_structure))[1])
    assert len(result.index) == ref_length


def test_align_to_ref_fail_if_bad_extension(
    structure_path_with_bad_extension: Path,
    target_structures: dict[str, Path],
    cluster: LocalCluster,
):
    with Client(cluster) as client:
        with pytest.raises(ValueError, match="TMalign only accepts the extensions:"):
            align_to_reference(structure_path_with_bad_extension, target_structures, client)



from __future__ import annotations
import json
import logging
from pathlib import Path

import attrs
import cattrs
import click
import pandas as pd
from biotite.sequence.seqtypes import ProteinSequence
from biotite.structure.residues import get_residues
from dask.base import compute
from dask.delayed import Delayed, delayed
from dask.distributed import Client, LocalCluster
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session

import apb.database.collection as collection
from apb.database.collection.interface import (
    add_preprocessed_profiles,
    get_all_protein_names,
    preprocess_profile,
)
from apb.profile import ProteinProfile
from apb.structure.convert import load_structure
from apb.structure.features.accessibility import calc_solvent_accessibility
from apb.structure.features.accessibility.datatypes import MaxSASASource, SolventAccessibilityConfig
from apb.structure.features.secondary_structure import calc_3_state_secondary_structure
from apb.structure.utils import find_structure_paths, structure_name_and_extension
from apb.types import Pathish, not_none
from apb.utils import converter

BATCH_SIZE: int = 1000


def save_feature_config(config: FeatureCalculationConfig, path: Pathish) -> Path:
    """Save a FeatureRecipe to file

    Note:
        - This function is not a method of `FeatureCalculationConfig` due to Dask
          parallelization requirements. Dask needs all involved objects to be
          serializable, and `FeatureCalculationConfig` is one of those objects.
          Instances of `FeatureCalculationConfig` can be converted to JSON using a
          cattrs `converter` object, as shown in this function. However, Dask
          serializes not just the instances but also the class itself. This poses a
          problem because the `converter` object is not easily serializable.

          If any methods in `FeatureCalculationConfig` reference the `converter` object,
          the class becomes non-serializable. The following attempts to serialize the
          `converter` object with Dask were unsuccessful:

            - Registering a cloudpickle patch for lru_cache objects, which led to
              downstream pickle issues.
              (https://github.com/cloudpipe/cloudpickle/issues/178#issuecomment-975735397)

            - Implementing a dill serialization family for Dask, which allowed the
              converter to be serialized but resulted in a 'TypeError: cannot pickle
              'KeyedRef' object'.
              (https://distributed.dask.org/en/stable/serialization.html)
    """
    with open(path, "w") as fp:
        json.dump(converter.unstructure(config), fp, indent=2)

    return Path(path)


def load_feature_config(path: Pathish) -> FeatureCalculationConfig:
    """Load a FeatureRecipe JSON from file

    Note:
        - See `save_feature_config` notes
    """
    with open(path) as fp:
        payload = json.load(fp)
        try:
            return converter.structure(payload, FeatureCalculationConfig)
        except cattrs.errors.ClassValidationError as e:
            user_message = (
                "The JSON file could not be structured as a feature config. Please "
                "check the input data and try again."
            )
            raise ValueError(user_message) from e


@attrs.define
class FeatureCalculationConfig:
    """Config that describes protein feature specifications

    This class holds all the tunables for calculating features like solvent
    accessibility, etc.

    Attributes:
        rsa_config: Configuration for relative solvent accessibility.
        sasa_config: Configuration for solvent accessible surface area.
    """

    rsa_config: SolventAccessibilityConfig
    sasa_config: SolventAccessibilityConfig

    @classmethod
    def default(cls) -> FeatureCalculationConfig:
        return cls(
            rsa_config=SolventAccessibilityConfig(
                normalize=True,
                source=MaxSASASource.WILKE,
                point_number=300,
            ),
            sasa_config=SolventAccessibilityConfig(
                normalize=False,
                source=MaxSASASource.WILKE,
                point_number=300,
            ),
        )


def _compute_and_insert(
    client: Client,
    engine: Engine,
    profiling_tasks: list[Delayed],
    batch_count: int,
) -> None:
    # Execute the profiling tasks in parallel
    profiles = compute(*profiling_tasks, scheduler=client)

    # Insert them (synchronously)
    with Session(engine) as session:
        add_preprocessed_profiles(session, profiles)

    logging.info(f"{BATCH_SIZE * batch_count + len(profiles)} profiles processed")


def build_profile(structure_path: Path, config: FeatureCalculationConfig) -> ProteinProfile:
    structure = load_structure(structure_path)
    sequence = ProteinSequence(get_residues(structure)[1])

    name = structure_name_and_extension(structure_path)[0] | not_none

    return ProteinProfile(
        name=name,
        sequence=sequence,
        structure=structure,
        residue_features=pd.DataFrame(
            {
                collection.Residue.secondary_structure.key: calc_3_state_secondary_structure(
                    structure
                ),
                collection.Residue.solvent_accessibility.key: calc_solvent_accessibility(
                    structure, config.sasa_config
                ),
                collection.Residue.relative_solvent_accessibility.key: calc_solvent_accessibility(
                    structure, config.rsa_config
                ),
            }
        ),
    )


def main(
    structure_dir: Pathish,
    db_path: Pathish,
    config: FeatureCalculationConfig,
    client: Client,
    batch_size: int = BATCH_SIZE,
) -> None:
    """Process structure files and add them to a ProteinDB.

    Args:
        directory: The directory containing structure files.
        db_path: Path to the database file.
        config: A feature config that is used to calculate features.
        batch_size:
            The DB is written in batches of protein profiles. This parameter controls
            the number of protein profiles held in the batch cache before they are
            inserted into the DB and released from memory.
    """
    logging.info(f"Using run config: {config}")

    engine = collection.get_engine(db_path, create=True)

    with Session(engine) as session:
        names = set(get_all_protein_names(session))

    profiling_tasks: list[Delayed] = []

    batch_count = 0
    duplicate_names = set()
    for structure_path in find_structure_paths(structure_dir):
        name = structure_name_and_extension(structure_path)[0] | not_none

        if name in names:
            # Skip proteins with names that have already been processed
            duplicate_names.add(name)
            continue

        profile_task = delayed(build_profile)(structure_path, config)
        preprocess_task = delayed(preprocess_profile)(profile_task)
        profiling_tasks.append(preprocess_task)

        names.add(name)

        if len(profiling_tasks) >= batch_size:
            _compute_and_insert(client, engine, profiling_tasks, batch_count)
            profiling_tasks.clear()

            batch_count += 1

    # Handle remaining tasks
    _compute_and_insert(client, engine, profiling_tasks, batch_count)
    profiling_tasks.clear()

    if len(duplicate_names):
        duplicate_list = list(duplicate_names)
        display_limit = min(len(duplicate_list), 100)
        logging.info(
            f"There were {len(duplicate_names)} duplicate proteins that were skipped. Here are "
            f"the first {display_limit}: {duplicate_list[:display_limit]}"
        )


@click.group(invoke_without_command=True)
@click.pass_context
def cli(ctx: click.Context) -> None:
    """Create a protein collection database

    A prototype CLI for creating a protein database that stores the underlying sequence,
    structure, residue features, and paired residue features for each entry.
    """
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


@cli.command()
@click.argument("path", type=click.Path())
def config(path: Path):
    """Create a feature config file and exit.

    This is useful if you want to control what features are calculated and how. FIXME
    this is mostly a placeholder for future development.
    """
    path = Path(path)
    if path.exists():
        raise FileExistsError(f"{path} already exists. Won't overwrite")

    save_feature_config(FeatureCalculationConfig.default(), path)
    click.echo(f"Feature config file created at {path}")


@cli.command()
@click.option(
    "-i",
    "--input",
    required=True,
    type=click.Path(exists=True),
    help="Directory containing structure files.",
)
@click.option(
    "-o",
    "--output",
    required=True,
    type=click.Path(),
    help="Path to the database file.",
)
@click.option(
    "-c",
    "--feature-config",
    required=False,
    type=click.Path(),
    help=(
        "Path to the configurable config file. Default will be used if not provided. Use `config` "
        "subcommand to generate a config file to use."
    ),
)
def run(input: Path, output: Path, feature_config: Path):
    """Load and process protein structures into a database."""
    config = (
        load_feature_config(feature_config)
        if feature_config
        else FeatureCalculationConfig.default()
    )
    with LocalCluster() as cluster, Client(cluster) as client:
        logging.info(f"Dask client: {client}")
        logging.info(f"View status at {client.dashboard_link}")
        main(input, output, config, client)


if __name__ == "__main__":
    cli()



import numpy as np
import pandas as pd
import pytest
from biotite.sequence.seqtypes import ProteinSequence
from biotite.structure.residues import get_residues

from apb._test_pointer import structure_paths
from apb.profile import ProteinProfile, get_residue_index
from apb.structure.convert import load_structure
from apb.types import ProteinStructure

STRUCTURE_PATH = structure_paths()[0]

np.random.seed(42)


@pytest.fixture
def structure() -> ProteinStructure:
    return load_structure(STRUCTURE_PATH)


@pytest.fixture
def sequence(structure: ProteinStructure) -> ProteinSequence:
    _, residues = get_residues(structure)
    return ProteinSequence(residues)


def _diverse_res_features(N: int) -> pd.DataFrame:
    int_series = pd.Series(np.random.randint(1, 100, size=N), name="Integer")
    float_series = pd.Series(np.random.random(N), name="Float")
    str_series = pd.Series([f"Item_{i}" for i in range(N)], name="String")
    bool_series = pd.Series(np.random.choice([True, False], size=N), name="Boolean")
    categorical_series = pd.Series(
        pd.Categorical(["c1", "c2", "3"][i % 3] for i in range(N)),
        name="Category",
    )

    return pd.DataFrame(
        {
            int_series.name: int_series,
            float_series.name: float_series,
            str_series.name: str_series,
            bool_series.name: bool_series,
            categorical_series.name: categorical_series,
        }
    )


def test_no_res_features(structure, sequence):
    # Create a featureless profile
    profile = ProteinProfile("test_profile", sequence, structure)

    # The amino acid column was absent, so it was added and equals the sequence
    assert "amino_acid" in profile.residue_features.columns
    assert "".join(sequence.symbols.tolist()) == "".join(
        profile.residue_features["amino_acid"].tolist()
    )

    # It has the proper index
    assert len(profile.residue_features.index) == len(sequence)
    assert profile.residue_features.index.equals(pd.Index(range(len(sequence))))


def test_yes_res_features(structure, sequence):
    res_features = _diverse_res_features(len(sequence))
    expected_index = get_residue_index(len(sequence))

    # The indices match so there's no problem
    assert res_features.index.equals(expected_index)
    profile = ProteinProfile("test_profile", sequence, structure, res_features)

    # The res_features frame _is_ the passed frame
    assert profile.residue_features is res_features

    # The amino acid column was absent, so it was added and equals the sequence
    assert "amino_acid" in profile.residue_features.columns
    assert "".join(sequence.symbols.tolist()) == "".join(
        profile.residue_features["amino_acid"].tolist()
    )

    # When the indices don't match, problem
    res_features.index = res_features.index.astype(str)
    assert not res_features.index.equals(expected_index)
    with pytest.raises(ValueError, match="Dataframe indices don't match expectation"):
        ProteinProfile("test_profile", sequence, structure, res_features)


def test_add_feature(structure, sequence):
    # Create a featureless profile
    profile = ProteinProfile("test_profile", sequence, structure)

    # Feature must be named
    with pytest.raises(ValueError, match="Unnamed series feature."):
        profile.add_feature(pd.Series(np.random.rand(len(sequence))))

    # Feature must have right length
    with pytest.raises(ValueError, match="must match sequence length"):
        profile.add_feature(pd.Series(np.random.rand(len(sequence) - 1), name="f1"))

    # Feature indices must match!
    with pytest.raises(ValueError, match="Dataframe indices don't match expectation"):
        feature = pd.Series(np.random.rand(len(sequence)), name="f1")
        feature.index += 1
        profile.add_feature(feature)

    # Signed 64-bit integer dtype is not the expected index
    feature = pd.Series(np.random.rand(len(sequence)), name="f1")
    feature.index = feature.index.astype(np.int64)
    assert feature.index.dtype != profile.residue_features.index.dtype

    # But datatypes can be coerced without altering res_features index
    index_before = profile.residue_features.index.copy(deep=True)
    profile.add_feature(feature)
    assert str(profile.residue_features.index.dtype) == str(index_before.dtype) == "uint32"
    assert not profile.residue_features[feature.name].isnull().any()  # type: ignore

    # Cannot add existing feature name
    with pytest.raises(ValueError, match="already present"):
        profile.add_feature(feature)



from pathlib import Path

import pytest

from apb._test_pointer import structure_paths
from apb.structure.convert import (
    load_structure,
    save_structure,
    structure_from_bytes,
    structure_to_bytes,
)
from apb.structure.utils import structures_almost_equal
from apb.types import ProteinStructure
from apb.utils import maybe_temp_directory

STRUCTURE_PATHS = structure_paths()
EXTENSIONS = [".pdb", ".pdb.gz", ".cif", ".bcif"]


@pytest.fixture
def structure() -> ProteinStructure:
    return load_structure(STRUCTURE_PATHS[0])


@pytest.mark.parametrize("path", STRUCTURE_PATHS)
def test_byte_round_trip(path: Path):
    # Load a structure
    structure = load_structure(path)

    # Represent it as bytes
    as_bytes = structure_to_bytes(structure)
    assert isinstance(as_bytes, bytes)

    # Now read it from bytes
    from_bytes = structure_from_bytes(as_bytes)

    # Besides floating point imprecision, it's an isomorphic round trip
    assert structures_almost_equal(structure, from_bytes)


@pytest.mark.parametrize("extension", EXTENSIONS)
def test_save_load_roundtrip(structure: ProteinStructure, extension: str):
    with maybe_temp_directory() as tmp_dir:
        path = tmp_dir / f"test{extension}"

        # Save the structure
        save_structure(path, structure)

        # It is in the file system
        assert path.exists()

        # Now load it
        structure_from_file = load_structure(path)

        # It equates
        assert structures_almost_equal(structure, structure_from_file)



from __future__ import annotations

import attrs
import pandas as pd
from biotite.sequence.seqtypes import ProteinSequence

from apb.types import ProteinStructure, aa_dtype


def get_residue_index(sequence_length: int) -> pd.Index:
    return pd.Index(range(sequence_length), "uint32", name="residue_index")


@attrs.define
class ProteinProfile:
    """A protein profile dataclass

    A protein profile is data structure that integrates sequence, structure, and
    associated annotation data for an individual protein. By tying sequence and
    structure components together within a protein profile, downstream analyses that
    incorporate information from both sequence and structure (very common) is made
    easier.

    Attributes:
        name: The name of the protein.
        sequence: The amino acid sequence of the protein.
        structure: The 3D structure of the protein.
        res_features:
            A dataframe containing residue-level features, indexed by residue_index. If
            not provided, a dataframe with a single column, "amino_acid", is used.
        source:
            The metadata describing the source of the protein. Currently this is a
            placeholder empty string but could be structured into something in the
            future.
    """

    name: str
    sequence: ProteinSequence
    structure: ProteinStructure
    residue_features: pd.DataFrame = attrs.field(factory=pd.DataFrame)
    source: str = attrs.field(default="")

    def __attrs_post_init__(self) -> None:
        expected_index = get_residue_index(len(self.sequence))

        if self.residue_features.empty:
            # Empty dataframe can still have index of non-zero length
            self.residue_features.index = expected_index
        else:
            # A non-empty dataframe was passed. Ensure the shape matches
            if self.residue_features.shape[0] != len(self.sequence):
                raise ValueError(
                    f"Expected axis 0 shape to match sequence length of {len(self.sequence)}, "
                    f"but got shape {self.residue_features.shape}."
                )

        # Ensure element-by-element consistency of indices
        if not self.residue_features.index.equals(expected_index):
            raise ValueError("Dataframe indices don't match expectation (range index from 0)")

        self.residue_features.index = expected_index

        if "amino_acid" not in self.residue_features:
            self.residue_features.insert(
                0, "amino_acid", pd.Series(self.sequence.symbols, dtype=aa_dtype)
            )

    def add_feature(self, feature: pd.Series) -> None:
        """Add a new residue-level feature

        Args:
            feature:
                A pandas Series representing the new feature, with a length matching the
                protein sequence.

        Raises:
            ValueError:
                If the length of the feature does not match the sequence length, if the
                feature is unnamed, if the feature name already exists, or if the
                feature index does not match the expected index (a range from 0 ->
                sequence length).
        """
        if len(feature) != len(self.sequence):
            raise ValueError(
                f"Feature of length {len(feature)} must match sequence length {len(self.sequence)}"
            )

        if feature.name is None:
            raise ValueError("Unnamed series feature. Add a unique name")

        if feature.name in self.residue_features.columns:
            raise ValueError(f"Feature '{feature.name}' already present")

        if not feature.index.equals(self.residue_features.index):
            raise ValueError("Dataframe indices don't match expectation (range index from 0)")

        feature.index = self.residue_features.index

        self.residue_features[feature.name] = feature



from __future__ import annotations

import attrs

from apb.vendor.strenum import StrEnum, auto


class MaxSASASource(StrEnum):
    """Different definitions of per-residue max exposed surface area

    Attributes
        AHMAD: Ahmad et al. 2003 https://doi.org/10.1002/prot.10328
        MILLER: Miller et al. 1987 https://doi.org/10.1016/0022-2836(87)90038-6
        SANDER: Sander & Rost 1994 https://doi.org/10.1002/prot.340200303
        WILKE: Tien et al. 2013 https://doi.org/10.1371/journal.pone.0080635
    """

    AHMAD = auto()
    MILLER = auto()
    SANDER = auto()
    WILKE = auto()


@attrs.define(frozen=True)
class SolventAccessibilityConfig:
    normalize: bool
    source: MaxSASASource
    point_number: int

    @classmethod
    def default(cls) -> SolventAccessibilityConfig:
        return cls(
            normalize=True,
            source=MaxSASASource.WILKE,
            point_number=200,
        )

    @property
    def feature_name(self) -> str:
        return "relative_solvent_accessibility" if self.normalize else "solvent_accessibility"



from itertools import combinations

import attrs
import numpy as np
import pandas as pd
import pytest
from biotite.sequence.seqtypes import ProteinSequence
from biotite.structure.residues import get_residues

from apb._test_pointer import structure_paths
from apb.structure.convert import load_structure
from apb.structure.features.accessibility import calc_solvent_accessibility
from apb.structure.features.accessibility.datatypes import MaxSASASource, SolventAccessibilityConfig

STRUCTURE_PATHS = structure_paths()
SASA_SOURCES = [source for source in MaxSASASource]

TEST_CONFIG = SolventAccessibilityConfig.default()


@pytest.mark.parametrize("path", STRUCTURE_PATHS)
@pytest.mark.parametrize("source", MaxSASASource)
def test_output(path, source):
    structure = load_structure(path)
    sequence = ProteinSequence(get_residues(structure)[1])
    config = attrs.evolve(TEST_CONFIG, source=source, normalize=True)
    out = calc_solvent_accessibility(structure, config)

    # Dtype is right
    assert out.dtype == "float64", "Values should be categorical dtype"

    # Name is correct
    assert out.name == "relative_solvent_accessibility"

    # Length matches sequence
    assert len(sequence) == len(out)

    # There are no null values
    assert not out.isnull().any()

    # Assert indices are properly ordered and start at 0
    assert out.index.equals(pd.Index(range(len(sequence))))


@pytest.mark.parametrize("path", STRUCTURE_PATHS)
@pytest.mark.parametrize("source", MaxSASASource)
def test_normalize(path, source):
    structure = load_structure(path)

    # Make a config with the specified source
    config = attrs.evolve(TEST_CONFIG, source=source, normalize=True)

    out = calc_solvent_accessibility(structure, config)

    # Range is between 0 and 1
    assert (out <= 1.0).all()
    assert (out >= 0).all()

    # Name is relative_solvent_accessibility
    assert out.name == "relative_solvent_accessibility"

    # Make an unnormalized config
    config = attrs.evolve(TEST_CONFIG, source=source, normalize=False)

    out = calc_solvent_accessibility(structure, config)

    # Values exceed 1.0
    assert (out > 1.0).any()
    assert (out >= 0).all()

    # Name is solvent_accessibility
    assert out.name == "solvent_accessibility"


@pytest.mark.parametrize("path", STRUCTURE_PATHS)
def test_sources(path):
    config = TEST_CONFIG
    structure = load_structure(path)

    outs = {}

    for source in MaxSASASource:
        # Make a config with the specified source
        config = attrs.evolve(TEST_CONFIG, source=source, normalize=True)

        outs[source] = calc_solvent_accessibility(structure, config).values

    # Ensure all outputs are different
    for source1, source2 in combinations(outs.keys(), 2):
        assert not np.array_equal(outs[source1], outs[source2]), (
            f"Outputs for sources {source1} and {source2} should not be identical"
        )



"""Points to test files present in development

Our test files live outside the python package so they aren't inadvertently linted.

In production, none of these paths should exist. In development, all of them
should.
"""

from pathlib import Path

TEST_ROOT = Path(__file__).parent.parent / "test_data"

TEST_STRUCTURES_ROOT = TEST_ROOT / "structures"

TEST_RAW_COLLECTION = TEST_ROOT / "raw_collection"


def structure_paths() -> list[Path]:
    return sorted(list(TEST_STRUCTURES_ROOT.glob("*")))


TEST_TMALIGN_ROOT = TEST_ROOT / "tmalign"



from __future__ import annotations
from collections.abc import Mapping
from functools import partial

import attrs
import numpy as np
import pandas as pd
from biotite.sequence.align.cigar import read_alignment_from_cigar
from biotite.sequence.seqtypes import ProteinSequence
from numpy.typing import NDArray

from apb.application.foldseek.base import (
    DEFAULT_CONVERTALIS_CONFIG,
    DEFAULT_SEARCH_CONFIG,
    convertalis,
    createdb,
    search,
)
from apb.application.foldseek.datatypes import FoldSeekConvertalisConfig, FoldSeekSearchConfig
from apb.application.foldseek.utils import create_tsv_for_createdb
from apb.types import Pathish, StructuralReferenceAligner
from apb.utils import maybe_temp_directory

DEFAULT_REFERENCE_ALIGNMENT_SEARCH_CONFIG = attrs.evolve(DEFAULT_SEARCH_CONFIG, backtrace=True)
DEFAULT_REFERENCE_ALIGNMENT_CONVERTALIS_CONFIG = attrs.evolve(DEFAULT_CONVERTALIS_CONFIG)
CIGAR_OUTPUT_FORMAT = "query,target,qstart,tstart,qend,tend,qseq,tseq,cigar"


def _ensure_valid_search_config(config: FoldSeekSearchConfig) -> None:
    if not config.backtrace:
        raise ValueError("Cannot pass search config with backtrace=False.")


def _generate_cigar_strings(
    query_path: Pathish,
    target_path: Pathish,
    search_config: FoldSeekSearchConfig,
    convertalis_config: FoldSeekConvertalisConfig,
    working_dir: Pathish | None = None,
) -> pd.DataFrame:
    """Generates CIGAR strings for the alignments.

    This runs `foldseek search` followed by `foldseek convertalis` with the output
    format defined by `CIGAR_OUTPUT_FORMAT.

    Returns:
        pd.DataFrame: A dataframe of the resultant hits table.
    """
    convertalis_config = attrs.evolve(convertalis_config, format_output=CIGAR_OUTPUT_FORMAT)

    with maybe_temp_directory(working_dir) as working_dir:
        alignment_path = working_dir / "aln"
        output_path = working_dir / "output.tsv"

        search(query_path, target_path, alignment_path, search_config)

        hit_table = convertalis(
            query_path=query_path,
            target_path=target_path,
            alignment_path_prefix=alignment_path,
            tsv_path=output_path,
            config=convertalis_config,
            working_dir=working_dir,
        )

    return hit_table


def _process_cigar_strings(cigars: pd.DataFrame) -> pd.DataFrame:
    """Creates an alignment trace from a cigar string hit table

    This accumulates pairwise alignment traces for each hit to the reference. The final
    result is compiled into a dataframe, where the index length equals the reference
    sequence length, and each column holds the trace values for the targets that aligned
    to the reference.

    Returns:
        pd.DataFrame: A dataframe of the alignment trace.
    """
    if cigars.empty:
        raise ValueError("There are 0 hits to align.")

    # The reference sequence is in the qseq column. Grab its length from the first entry.
    ref_length = len(cigars.iloc[0].qseq)

    traces: dict[str, NDArray[np.int64]] = {}
    for _, hit in cigars.iterrows():
        # Foldseek outputs cigar strings that consume the target sequence beginning
        # from the alignment start. That means we have to splice the reference and
        # target sequences to their respective aligned segments.
        target_aligned_sequence = ProteinSequence(hit.tseq)[hit.tstart - 1 : hit.tend]
        reference_aligned_sequence = ProteinSequence(hit.qseq)[hit.qstart - 1 : hit.qend]

        # Create the alignment from the cigar string and the sequences. Confusingly,
        # we pass the target sequence to the argument `reference_sequence` and the
        # reference sequence to the argument `segment_sequence`. This is on purpose
        # and the result of biotite's opinionated nomenclature.
        alignment = read_alignment_from_cigar(
            cigar=hit.cigar,
            position=0,
            reference_sequence=target_aligned_sequence,
            segment_sequence=reference_aligned_sequence,
        )

        target_trace = alignment.trace[:, 0]
        reference_trace = alignment.trace[:, 1]

        # Since the alignment can start anywhere along the two sequences, we need to
        # offset all non-gapped (-1) trace values by the alignment start index. Note
        # that it would have sufficed to do this only for the target sequence by
        # supplying `hit.tstart - 1` to the `position` argument above (in
        # `read_alignment_from_cigar`), but for posterity we do both the target and
        # reference sequence together here:
        target_trace[target_trace > -1] += hit.tstart - 1
        reference_trace[reference_trace > -1] += hit.qstart - 1

        # Now remove all align indices in the target where the reference is gapped.
        target_trace = target_trace[reference_trace != -1]

        # The trace is currently only as long as the alignment, which varies from
        # hit to hit. To be able to compare all the traces with one another, we pad
        # each trace with the appropriate number of N-terminus and C-terminus gap
        # values (-1) such that each trace length equals the length of the
        # reference.
        n_terminal_gaps = -np.ones(hit.qstart - 1, dtype=np.int64)
        c_terminal_gaps = -np.ones(ref_length - hit.qend, dtype=np.int64)
        target_trace = np.concatenate((n_terminal_gaps, target_trace, c_terminal_gaps))

        if len(target_trace) != ref_length:
            raise RuntimeError(
                f"After concatenating N- and C-terminal gaps, the target ({hit.target}) trace "
                f"length ({len(target_trace)}) must equal the reference ({hit.query}) length "
                f"({ref_length}). The number of N- and C- terminal gap characters to pad was "
                f"calculated to be {len(n_terminal_gaps)} and {len(c_terminal_gaps)}, "
                f"respectively. Please report this issue."
            )

        traces[hit.target] = target_trace

    return pd.DataFrame(traces)


def align_to_reference(
    reference_structure: Pathish,
    target_structures: Mapping[str, Pathish],
    search_config: FoldSeekSearchConfig = DEFAULT_REFERENCE_ALIGNMENT_SEARCH_CONFIG,
    convertalis_config: FoldSeekConvertalisConfig = DEFAULT_REFERENCE_ALIGNMENT_CONVERTALIS_CONFIG,
):
    """Performs alignment of a collection of structures to a reference structure using Foldseek.

    Executes a pairwise alignment to the reference structure for each target structure
    and returns the alignment trace.

    Args:
        reference_structure: Path to the reference structure file.
        target_structures: Dictionary with keys as names and values as structure paths.
        search_config:
            Configuration for the Foldseek search. If providing this value, `backtrace`
            must be set to true, which is required for the alignment.
        convertalis_config:
            Configuration for foldseek's `convertalis` subcommand. The format_output
            attribute is ignored and overwritten with `CIGAR_OUTPUT_FORMAT`.

    Returns:
        pd.DataFrame:
            A dataframe of the alignment trace. Each column is a target hit to the
            reference. The indices are the reference residue indices. For further
            details, see `apb.utils.StructuralReferenceAligner`.

    Raises:
        ValueError: If no hits are found in the alignment process.
    """
    _ensure_valid_search_config(search_config)

    with maybe_temp_directory() as working_dir:
        query_path = working_dir / "query"
        target_path = working_dir / "target"
        target_input = working_dir / "createdb_input.tsv"

        create_tsv_for_createdb(target_input, target_structures.values())
        createdb(reference_structure, query_path)
        createdb(target_input, target_path)

        cigars = _generate_cigar_strings(
            query_path,
            target_path,
            search_config,
            convertalis_config,
            working_dir,
        )

        return _process_cigar_strings(cigars)


def create_reference_aligner(
    search_config: FoldSeekSearchConfig = DEFAULT_REFERENCE_ALIGNMENT_SEARCH_CONFIG,
    convertalis_config: FoldSeekConvertalisConfig = DEFAULT_REFERENCE_ALIGNMENT_CONVERTALIS_CONFIG,
) -> StructuralReferenceAligner:
    """Creates a structural reference aligner.

    This function configures and returns a partially applied version of
    `align_to_reference` with predefined search and convertalis configurations.

    Args:
        search_config: Configuration for the Foldseek's `search` subcommand..
        convertalis_config: Configuration for Foldseek's `convertalis` subcommand.

    Returns:
        StructuralReferenceAligner:
            A function that performs structure alignment using the specified
            configurations. See `apb.utils.StructuralReferenceAligner` for details about
            the function's inputs and outputs.
    """
    return partial(
        align_to_reference,
        search_config=search_config,
        convertalis_config=convertalis_config,
    )



from pathlib import Path

import numpy as np
import pytest

from apb._test_pointer import TEST_RAW_COLLECTION
from apb.application.foldseek.align import (
    CIGAR_OUTPUT_FORMAT,
    DEFAULT_REFERENCE_ALIGNMENT_CONVERTALIS_CONFIG,
    DEFAULT_REFERENCE_ALIGNMENT_SEARCH_CONFIG,
    _ensure_valid_search_config,
    _generate_cigar_strings,
    align_to_reference,
)
from apb.application.foldseek.base import createdb
from apb.application.foldseek.datatypes import FoldSeekSearchConfig
from apb.application.foldseek.utils import create_tsv_for_createdb
from apb.structure.utils import structure_name_and_extension
from apb.types import not_none
from apb.utils import maybe_temp_directory


@pytest.fixture
def db(tmp_path: Path) -> Path:
    return createdb(TEST_RAW_COLLECTION, tmp_path / "output_db")


@pytest.fixture
def reference_path() -> Path:
    return TEST_RAW_COLLECTION / "A0A0A7RVV1.pdb.gz"


@pytest.fixture
def target_structures() -> dict[str, Path]:
    targets = {}
    for path in sorted(list(TEST_RAW_COLLECTION.glob("*"))):
        name = structure_name_and_extension(path)[0] | not_none
        targets[name] = path

    return targets


def test_ensure_valid_search_config():
    valid_config = FoldSeekSearchConfig(backtrace=True)
    _ensure_valid_search_config(valid_config)

    invalid_config = FoldSeekSearchConfig(backtrace=False)
    with pytest.raises(ValueError, match="Cannot pass search config with backtrace=False."):
        _ensure_valid_search_config(invalid_config)


def test_generate_cigar_strings(reference_path: Path, target_structures: dict[str, Path]):
    with maybe_temp_directory() as working_dir:
        query_path = working_dir / "query"
        target_path = working_dir / "target"
        target_input = working_dir / "createdb_input.tsv"

        create_tsv_for_createdb(target_input, target_structures.values())
        createdb(reference_path, query_path)
        createdb(target_input, target_path)

        cigars = _generate_cigar_strings(
            query_path,
            target_path,
            DEFAULT_REFERENCE_ALIGNMENT_SEARCH_CONFIG,
            DEFAULT_REFERENCE_ALIGNMENT_CONVERTALIS_CONFIG,
            working_dir,
        )

    # The cigars frame is typical convertalis output
    assert not cigars.empty
    assert cigars.columns.tolist() == CIGAR_OUTPUT_FORMAT.split(",")

    # Reference has a fully-matching cigar string with itself
    reference_name = structure_name_and_extension(reference_path)[0] | not_none
    self_hit = cigars.query(f"target == '{reference_name}'").iloc[0]
    expected_cigar = f"{len(self_hit.qseq)}M"
    assert self_hit.cigar == expected_cigar

    # Check that one of the cigar strings matches the expected value (the 7th row was
    # chosen at random and was manually validated)
    assert cigars.iloc[7].cigar == "312M10D64M"


def test_align_to_reference(reference_path: Path, target_structures: dict[str, Path]):
    traces = align_to_reference(
        reference_structure=reference_path,
        target_structures=target_structures,
        search_config=DEFAULT_REFERENCE_ALIGNMENT_SEARCH_CONFIG,
        convertalis_config=DEFAULT_REFERENCE_ALIGNMENT_CONVERTALIS_CONFIG,
    )

    reference_name = structure_name_and_extension(reference_path)[0] | not_none
    reference_length = traces.shape[0]

    # The number of trace rows matches the reference
    assert traces.shape[0] == reference_length

    # The self-trace is an arange that equals the index
    assert (traces[reference_name].values == traces.index.values).all()  # type: ignore
    assert (traces[reference_name].values == np.arange(reference_length)).all()  # type: ignore



from collections.abc import Mapping
from pathlib import Path
from typing import Protocol, TypeVar, overload

import pandas as pd
from biotite.structure.atoms import AtomArray, AtomArrayStack

ProteinStructure = AtomArray | AtomArrayStack

Pathish = str | Path


class StructuralReferenceAligner(Protocol):
    """A structural aligner protocol.

    This class defines a call signature that functions that align proteins to a
    reference are expected to comply with. There is intentionally no implementation.
    """

    def __call__(
        self, reference_structure: Pathish, target_structures: Mapping[str, Pathish]
    ) -> pd.DataFrame:
        """Per-residue alignment of proteins to a reference based on structural similarity.

        Returns an alignment trace that mimics the datastructure used in biotite:

            "The trace is a (m x n) ndarray with alignment length m and sequence count n.
            Each element of the trace is the index in the corresponding sequence. A gap is
            represented by the value -1."
                - (https://www.biotite-python.org/apidoc/biotite.sequence.align.Alignment.html)

        For example, the following alignment:

        CGTCAT--
        --TCATGC

        Has the following alignment trace:

        [[ 0 -1]
         [ 1 -1]
         [ 2  0]
         [ 3  1]
         [ 4  2]
         [ 5  3]
         [-1  4]
         [-1  5]]

        Args:
            reference_structure: Path to the reference structure file.
            target_structures: Dictionary with keys as names and values as structure paths.

        Returns:
            pd.DataFrame:
                A dataframe of the alignment trace. Each column is a target hit to the
                reference. The indices are the reference residue indices.
        """
        ...


_amino_acids: list[str] = [
    "A",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "I",
    "K",
    "L",
    "M",
    "N",
    "P",
    "Q",
    "R",
    "S",
    "T",
    "V",
    "W",
    "Y",
]

_amino_acids_gapped = _amino_acids + ["-"]

# The 20 canonical amino acids
aa_dtype = pd.CategoricalDtype(categories=_amino_acids)

# The 20 canonical amino acids and the gap character
aa_gapped_dtype = pd.CategoricalDtype(categories=_amino_acids_gapped)

# Secondary structure (3-state): [a]lpha helix, [b]eta sheet, [c]oil
ss3_dtype = pd.CategoricalDtype(categories=["a", "b", "c"])


T = TypeVar("T")


class NotNone:
    """A utility class ensure values are not None in a type-safe manner.

    This class allows for a convenient way to ensure that a value is not None using the
    | operator or __call__. If the value is None, a ValueError is raised. This lets type
    checkers to infer that the value is not None after the operation.

    Examples:

        All three of these examples either prevent or ignore a type checking error, but
        the last method is preferred.

        1. Sloppy

        >>> ans = func_returns_str_or_none()
        >>> ans.split(" ")  # type: ignore

        - Error is ignored
        - The type of `ans` is ignored altogether

        2. Better

        >>> ans = func_returns_str_or_none()
        >>> assert ans is not None
        >>> ans.split(" ")

        - Explicit
        - `ans` is type-checked as a string
        - If run with python -O, this is not runtime safe
        - Verbose

        3. Best

        >>> ans = func_returns_str_or_none() | not_none
        >>> ans.split(" ")

        OR

        >>> ans = not_none(func_returns_str_or_none())
        >>> ans.split(" ")

        - Explicit
        - Runtime safe
        - Understood by type checkers
    """

    @overload
    def __ror__(self, other: T | None) -> T: ...

    @overload
    def __ror__(self, other: T) -> T: ...

    def __ror__(self, other: T | None) -> T:
        return self._not_none(other)

    def __call__(self, value: T | None) -> T:
        return self._not_none(value)

    def _not_none(self, obj: T | None) -> T:
        if obj is None:
            raise ValueError("Value cannot be None")
        return obj


not_none = NotNone()



from pathlib import Path

import pytest

from apb._test_pointer import structure_paths
from apb.structure.convert import (
    load_structure,
    save_structure,
    structure_from_bytes,
    structure_to_bytes,
)
from apb.structure.utils import (
    RECOGNIZED_STRUCTURE_FILETYPES,
    extension_matches,
    find_structure_paths,
    structure_name_and_extension,
    structures_almost_equal,
)
from apb.utils import maybe_temp_directory

STRUCTURE_PATHS = structure_paths()


def test_structure_extension():
    # Base extensions work
    assert structure_name_and_extension("example.pdb") == ("example", ".pdb")
    assert structure_name_and_extension("example.cif") == ("example", ".cif")
    assert structure_name_and_extension("example.bcif") == ("example", ".bcif")

    # Supported compound extensions work
    assert structure_name_and_extension("example.pdb.gz") == ("example", ".pdb.gz")

    # Basenames with dots are OK
    assert structure_name_and_extension("example.0.pdb") == ("example.0", ".pdb")
    assert structure_name_and_extension("example.0.pdb.gz") == ("example.0", ".pdb.gz")
    assert structure_name_and_extension("example.0.cif") == ("example.0", ".cif")
    assert structure_name_and_extension("example.0.bcif") == ("example.0", ".bcif")

    # Diabolical cases don't take the bait
    assert structure_name_and_extension("example.pdb.gz.example") == (None, None)
    assert structure_name_and_extension("example.pdb.example") == (None, None)

    # Unsupported returns None tuple
    assert structure_name_and_extension("example") == (None, None)
    assert structure_name_and_extension("example.pdb.zip") == (None, None)
    assert structure_name_and_extension("example.bcif.gz") == (None, None)
    assert structure_name_and_extension("example.cif.gz") == (None, None)
    assert structure_name_and_extension("example.mmcif") == (None, None)


def test_find_structure_files():
    with maybe_temp_directory() as root:
        subdir1 = root / "subdir1"
        subdir1.mkdir()

        subdir2 = root / "subdir2"
        subdir2.mkdir()

        # Create structure files
        structure_file0 = subdir1 / "file1.pdb.gz"
        structure_file0.touch()
        structure_file1 = subdir1 / "file1.pdb"
        structure_file1.touch()
        structure_file2 = subdir1 / "file2.bcif"
        structure_file2.touch()
        structure_file3 = subdir2 / "fil.e3.cif"
        structure_file3.touch()

        # Unsupported structure file
        structure_file4 = root / "file4.cif.gz"
        structure_file4.touch()

        # Create non-structure files
        non_structure_file1 = subdir1 / "file5.txt"
        non_structure_file1.touch()
        non_structure_file2 = subdir2 / "file6.doc"
        non_structure_file2.touch()
        non_structure_file3 = root / "file7"
        non_structure_file3.touch()

        # Exhaust the generator, collecting the paths and the summary
        generator = find_structure_paths(root)
        structure_files: list[Path] = []
        try:
            while True:
                structure_files.append(next(generator))
        except StopIteration as e:
            summary = e.value

        # Structure files are in the result
        assert structure_file0 in structure_files
        assert structure_file1 in structure_files
        assert structure_file2 in structure_files
        assert structure_file3 in structure_files

        # Unsupported structure files are not
        assert structure_file4 not in structure_files

        # Non-structure files are not
        assert non_structure_file1 not in structure_files
        assert non_structure_file2 not in structure_files
        assert non_structure_file3 not in structure_files

        # The summary matches expectation
        assert summary.num_hits == 4
        assert summary.num_misses == 4
        assert summary.hit_exts == {".pdb", ".bcif", ".cif", ".pdb.gz"}
        assert summary.miss_exts == {".txt", ".doc", "", ".gz"}


@pytest.mark.parametrize("extension", RECOGNIZED_STRUCTURE_FILETYPES)
def test_extension_matches(extension: str):
    path = f"test_structure{extension}"

    # Default, which is all recognized structure filetypes, always matches
    assert extension_matches(path)

    # Matches if the extension is in `allowed`
    allowed = {".pdb.gz", ".cif"}
    assert extension_matches(path, allowed) == (extension in allowed)


@pytest.mark.parametrize("path", STRUCTURE_PATHS)
def test_structures_almost_equal(path: Path):
    # Load the structure
    structure = load_structure(path)

    # Structure almost equals its copy
    assert structures_almost_equal(structure, structure.copy())

    # Write, then read from various formats
    with maybe_temp_directory() as tmp_dir:
        base = tmp_dir / "protein"

        for ext in RECOGNIZED_STRUCTURE_FILETYPES:
            path = base.with_suffix(ext)
            save_structure(path, structure)

            assert structures_almost_equal(structure, load_structure(path))

    # Byte transformations also OK
    from_bytes = structure_from_bytes(structure_to_bytes(structure))
    assert structures_almost_equal(structure, from_bytes)

    # Changing coords yields False
    structure_copy = structure.copy()
    assert structure_copy.coord is not None
    structure_copy.coord[0] *= 1.0001
    assert not structures_almost_equal(structure, structure_copy)

    # Adding annotations yields False
    structure_copy = structure.copy()
    structure_copy.add_annotation("foo", dtype=bool)
    assert not structures_almost_equal(structure, structure_copy)

    # Setting annotations yields False
    structure_copy = structure.copy()
    structure_copy.set_annotation("bar", list(range(len(structure_copy))))
    assert not structures_almost_equal(structure, structure_copy)



import logging
from collections.abc import Generator
from pathlib import Path

import attrs
import numpy as np

from apb.types import Pathish, ProteinStructure

EQUIVALENCY_TOLERANCE: float = 1e-5

RECOGNIZED_STRUCTURE_FILETYPES: set[str] = {".pdb", ".pdbx", ".pdb.gz", ".bcif", ".cif"}


def structure_name_and_extension(path: Pathish) -> tuple[str | None, str | None]:
    """Extract the base name and supported structure file extension of a path if it exists

    This only finds structure file extensions if they are supported for parsing. It
    includes commonly used compound extensions (.pdb.gz). For a list of supported
    extensions, see EXTENSION_SCOPE.

    Args:
        path: The file path from which to extract the base name and structure extension.

    Returns:
        tuple[str | None, str | None]:
            A tuple containing the base name and the extension if it is supported.
            If unsupported or unrecognized, returns (None, None).

    Examples:
        - For a path 'example.pdb', it returns ('example', '.pdb').
        - For a path 'example.pdb.gz', it returns ('example', '.pdb.gz').
        - For a path 'example', it returns (None, None).
    """
    path = Path(path)
    suffixes = path.suffixes

    combined_suffix = ""
    for suffix in reversed(suffixes):
        combined_suffix = suffix + combined_suffix
        if combined_suffix in RECOGNIZED_STRUCTURE_FILETYPES:
            base_name = path.name[: -len(combined_suffix)]
            return base_name, combined_suffix

    return None, None


def extension_matches(path: Pathish, match_set: set[str] = RECOGNIZED_STRUCTURE_FILETYPES) -> bool:
    """Returns whether the path has an extension matching the set of specified file extensions."""
    if not match_set.issubset(RECOGNIZED_STRUCTURE_FILETYPES):
        raise ValueError(f"Members of `match_set` must belong to {RECOGNIZED_STRUCTURE_FILETYPES}.")

    return structure_name_and_extension(path)[1] in match_set


@attrs.define
class _StructureFileSearchSummary:
    """Logging accumulator for `find_structure_paths`"""

    num_hits: int = attrs.field(default=0)
    num_misses: int = attrs.field(default=0)
    hit_exts: set[str] = attrs.field(factory=set)
    miss_exts: set[str] = attrs.field(factory=set)

    def add_hit(self, ext: str) -> None:
        self.num_hits += 1
        self.hit_exts.add(ext)

    def add_miss(self, ext: str) -> None:
        self.num_misses += 1
        self.miss_exts.add(ext)

    def summary(self) -> str:
        return (
            f"Number of structure files parsed: {self.num_hits}; "
            f"Number of non-structure files skipped: {self.num_misses}; "
            f"Unique structure file extensions: {self.hit_exts}; "
            f"Unique non-structure file extensions: {self.miss_exts}"
        )


def find_structure_paths(dir: Pathish) -> Generator[Path, None, _StructureFileSearchSummary]:
    summary = _StructureFileSearchSummary()

    dir = Path(dir)

    if not dir.is_dir():
        raise ValueError(f"{dir} is not a dir.")

    for path in dir.rglob("*"):
        if path.is_dir():
            continue

        _, valid_ext = structure_name_and_extension(path)

        if valid_ext is None:
            summary.add_miss("" if not len(path.suffix) else path.suffix)
            continue

        summary.add_hit(valid_ext)
        yield path

    logging.info(summary.summary())
    return summary


def structures_almost_equal(
    structure1: ProteinStructure,
    structure2: ProteinStructure,
    rtol: float = EQUIVALENCY_TOLERANCE,
) -> bool:
    """Compare two structures with tolerance for floating-point imprecision.

    Biotite implements a strict __eq__ method with no float precision leniency, so the
    `==` operator can yield False merely due to imprecisions introduced by isomorphic
    transformations and I/O operations.

    This function implements a looser comparison, allowing for a relative tolerance of
    array elements of `rtol`.

    Parameters:
        structure1: The first protein structure to compare.
        structure2: The second protein structure to compare.

    Returns:
        bool: True if the structures are considered equivalent, otherwise False.

    Notes:
        - For precise comparison, use `structure1 == structure2`
        - Implementation is borrowed from the essence of biotite structure __eq__ logic
          (from biotite/structure/atoms.py) and lower the standard by replacing
          array_equal calls with allclose calls.
    """

    # Biotite implements a strict __eq__ method
    if structure1 == structure2:
        return True

    # Check coordinates equivalence
    if structure1._coord is None or structure2._coord is None:
        return False
    if not np.allclose(structure1._coord, structure2._coord, rtol=rtol):
        return False

    if sorted(structure1._annot.keys()) != sorted(structure2._annot.keys()):
        return False

    # Check annotation values equivalence
    for name in structure1._annot:
        float_castable = True
        try:
            structure1._annot[name].astype(float)
            structure2._annot[name].astype(float)
        except ValueError:
            float_castable = False

        if float_castable:
            # Compare float-casted arrays
            if not np.allclose(
                structure1._annot[name].astype(float),
                structure2._annot[name].astype(float),
                rtol=rtol,
            ):
                return False
        else:
            if not np.array_equal(structure1._annot[name], structure2._annot[name]):
                return False

    if structure1._bonds != structure2._bonds:
        return False

    # Check box equivalence
    if structure1._box is None and structure2._box is None:
        pass  # Both are None, so equivalent in this context
    elif structure1._box is not None and structure2._box is not None:
        if not np.allclose(structure1._box, structure2._box, rtol=rtol):
            return False
    else:
        return False

    return True



from apb.structure.features.accessibility.core import calc_solvent_accessibility
from apb.structure.features.accessibility.datatypes import MaxSASASource, SolventAccessibilityConfig

__all__ = [
    "SolventAccessibilityConfig",
    "MaxSASASource",
    "calc_solvent_accessibility",
]
